import base64
from os import PathLike
from typing import cast

from anthropic.types import MessageParam

from mirascope.core import BaseMessageParam
from mirascope.core.anthropic._utils import convert_message_params
from mirascope.core.base import (
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)


class AnthropicMessageParamConverter(BaseMessageParamConverter):
    """Converts between Anthropic `MessageParam` objects and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[MessageParam]:
        """
        Convert from Mirascope `BaseMessageParam` to Anthropic's `MessageParam`.
        """
        return convert_message_params(
            cast(list[BaseMessageParam | MessageParam], message_params)
        )

    @staticmethod
    def from_provider(message_params: list[MessageParam]) -> list[BaseMessageParam]:
        """
        Convert from Anthropic's `MessageParam` back to Mirascope `BaseMessageParam`.
        """
        converted: list[BaseMessageParam] = []
        for message_param in message_params:
            content = message_param["content"]
            if isinstance(content, str):
                converted.append(
                    BaseMessageParam(role=message_param["role"], content=content)
                )
                continue
            converted_content = []

            for block in content:
                if not isinstance(block, dict):
                    continue

                if block["type"] == "text":
                    text = block.get("text")
                    if not isinstance(text, str):
                        raise ValueError(
                            "TextBlockParam must have a string 'text' field."
                        )
                    converted_content.append(TextPart(type="text", text=text))

                elif block["type"] == "image":
                    source = block.get("source")
                    source_type = source.get("type") if source else None
                    if not source or source_type not in ["base64", "url"]:
                        raise ValueError(
                            "ImageBlockParam must have a 'source' with type='base64' or type='url'."
                        )
                    if source_type == "url":
                        url = source.get("url")
                        if not url:
                            raise ValueError(
                                "ImageBlockParam source with type='url' must have a 'url'."
                            )
                        converted_content.append(
                            ImageURLPart(type="image_url", url=url, detail=None)
                        )
                    else:
                        image_data = source.get("data")
                        media_type = source.get("media_type")
                        if not image_data or not media_type:
                            raise ValueError(
                                "ImageBlockParam source with type='base64' must have 'data' and 'media_type'."
                            )
                        if media_type not in [
                            "image/jpeg",
                            "image/png",
                            "image/gif",
                            "image/webp",
                        ]:
                            raise ValueError(
                                f"Unsupported image media type: {media_type}. "
                                "BaseMessageParam currently only supports JPEG, PNG, GIF, and WebP images."
                            )
                        if isinstance(image_data, str):
                            decoded_image_data = base64.b64decode(image_data)
                        elif isinstance(image_data, PathLike):
                            with open(image_data, "rb") as image_data:
                                decoded_image_data = image_data.read()
                        else:
                            decoded_image_data = image_data.read()
                        converted_content.append(
                            ImagePart(
                                type="image",
                                media_type=media_type,
                                image=decoded_image_data,
                                detail=None,
                            )
                        )

                elif block["type"] == "tool_use":
                    converted_content.append(
                        ToolCallPart(
                            type="tool_call",
                            args=cast(dict, block["input"]),
                            id=block["id"],
                            name=block["name"],
                        )
                    )
                elif block["type"] == "tool_result":
                    converted_content.append(
                        ToolResultPart(
                            type="tool_result",
                            content=block["content"]  # pyright: ignore [reportTypedDictNotRequiredAccess]
                            if isinstance(block["content"], str)  # pyright: ignore [reportTypedDictNotRequiredAccess]
                            else list(block["content"])[0]["text"],  # pyright: ignore [reportTypedDictNotRequiredAccess, reportGeneralTypeIssues]
                            id=block["tool_use_id"],
                            is_error=block.get("is_error", False),
                        )
                    )
                else:
                    # Any other block type is not supported
                    raise ValueError(
                        f"Unsupported block type '{block['type']}'. "
                        "BaseMessageParam currently only supports text and image parts."
                    )
            if converted_content:
                converted.append(
                    BaseMessageParam(
                        role=message_param["role"], content=converted_content
                    )
                )

        return converted



from __future__ import annotations

import asyncio
import base64
import inspect
import json
import os
from collections import defaultdict
from collections.abc import Callable
from functools import lru_cache
from io import BytesIO
from typing import (
    Any,
    Literal,
    TypeAlias,
    TypedDict,
    TypeVar,
)

import websockets
from pydantic import BaseModel
from pydub import AudioSegment
from typing_extensions import NotRequired, overload
from websockets.asyncio.client import ClientConnection, connect

from ....core import BaseTool
from ....core.base._utils import (
    convert_base_model_to_base_tool,
    convert_function_to_base_tool,
)
from ._utils._audio import (
    async_audio_input_audio_buffer_append_event,
    async_audio_to_item_create_event,
    audio_chunk_to_audio_segment,
)
from ._utils._protocols import FunctionCallHandlerFunc, ReceiverFunc, SenderFunc
from .tool import FunctionCallArguments, OpenAIRealtimeTool, RealtimeToolParam

ChunkResponseType: TypeAlias = Literal[
    "text_chunk",
    "audio_chunk",
    "audio_transcript_chunk",
    "tool_chunk",
]
TextResponseType: TypeAlias = Literal[
    "text",
    "audio_transcript",
]
AudioResponseType: TypeAlias = Literal["audio"]
ResponseType: TypeAlias = ChunkResponseType | TextResponseType | AudioResponseType


_ResponseT = TypeVar("_ResponseT")
_ConnectionT = TypeVar("_ConnectionT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ToolSchemaT = TypeVar("_ToolSchemaT")

# TODO: Improve the type of response
Response: TypeAlias = Any
ToolName: TypeAlias = str
CallId: TypeAlias = str
ResponseId: TypeAlias = str
ToolType: TypeAlias = type[OpenAIRealtimeTool]


class RawMessage(TypedDict, total=False):
    event_id: str
    type: str
    item: NotRequired[dict]
    delta: NotRequired[str]
    response_id: NotRequired[str]
    tool_name: NotRequired[str]
    response: NotRequired[dict]
    name: NotRequired[str]
    call_id: NotRequired[str]


class Sender(TypedDict):
    func: SenderFunc[_ResponseT, OpenAIRealtimeTool]
    wait_for_text_response: bool
    wait_for_audio_transcript_response: bool
    tools: list[type[BaseModel] | Callable]


class FunctionCallHandler(TypedDict):
    tool_name: ToolName
    func: FunctionCallHandlerFunc


NotSet = object()


@lru_cache
def _convert_to_base_tool(tool: type[BaseModel] | Callable) -> type[ToolType]:
    if inspect.isclass(tool):
        convert_base_model_to_base_tool(tool, OpenAIRealtimeTool)
    return convert_function_to_base_tool(tool, OpenAIRealtimeTool)


def _get_tool_type_and_tool_schemas(
    tools: list[type[BaseModel] | Callable],
) -> dict[ToolName, tuple[ToolType, RealtimeToolParam]]:
    return {
        _tool_type._name(): (_tool_type, _tool_type.tool_schema())
        for tool in tools
        if (_tool_type := _convert_to_base_tool(tool))
    }


class Realtime:
    def __init__(
        self,
        model: str,
        *,
        context: dict[str, Any] | None = None,
        modalities: list[str] | NotSet = NotSet,
        instructions: str | NotSet = NotSet,
        voice: str | NotSet = NotSet,
        turn_detection: dict[str, Any] | None | NotSet = NotSet,
        input_audio_format: str | NotSet = NotSet,
        output_audio_format: str | NotSet = NotSet,
        input_audio_transcription: str | NotSet = NotSet,
        tool_choice: str | NotSet = NotSet,
        temperature: float | NotSet = NotSet,
        max_response_output_tokens: int | float | Literal["inf"] = NotSet,
        tools: list[type[BaseModel] | Callable] | NotSet = NotSet,
    ) -> None:
        self.senders: list[Sender] = []
        self.receivers: defaultdict[ResponseType, ReceiverFunc] = defaultdict(list)
        self.function_call_handlers: dict[ToolName, FunctionCallHandlerFunc] = {}
        self._connection: connect | None = None
        self.context = context or {}
        self._tool_schemas: dict[ToolName, tuple[ToolType, RealtimeToolParam]] = (
            _get_tool_type_and_tool_schemas(tools) if tools is not NotSet else {}
        )
        self._temporary_tool_schemas: defaultdict[
            ToolName, list[tuple[ToolType, RealtimeToolParam]]
        ] = defaultdict(list)
        self._session: dict[str, Any] = {
            "model": model,
            "modalities": modalities,
            "instructions": instructions,
            "voice": voice,
            "turn_detection": turn_detection,
            "input_audio_format": input_audio_format,
            "output_audio_format": output_audio_format,
            "input_audio_transcription": input_audio_transcription,
            "tool_choice": tool_choice,
            "temperature": temperature,
            "max_response_output_tokens": max_response_output_tokens,
            "tools": [tool_schema for _, tool_schema in self._tool_schemas.values()]
            if self._tool_schemas
            else NotSet,
        }
        self._text_message_received: bool = True
        self._audio_transcript_received: bool = True

    async def _wait_for_text_message(self) -> None:
        while not self._text_message_received:
            await asyncio.sleep(0.1)

    async def _wait_for_audio_transcript(self) -> None:
        while not self._audio_transcript_received:
            await asyncio.sleep(0.1)

    def sender(
        self,
        wait_for_text_response: bool = False,
        wait_for_audio_transcript_response: bool = False,
        tools: list[type[BaseModel] | Callable] | None = None,
    ) -> Callable[
        [SenderFunc[_ResponseT, OpenAIRealtimeTool]],
        SenderFunc[_ResponseT, OpenAIRealtimeTool],
    ]:
        def inner(
            func: SenderFunc[_ResponseT, OpenAIRealtimeTool],
        ) -> SenderFunc[_ResponseT, OpenAIRealtimeTool]:
            self.senders.append(
                {
                    "func": func,
                    "wait_for_text_response": wait_for_text_response,
                    "wait_for_audio_transcript_response": wait_for_audio_transcript_response,
                    "tools": tools,
                }
            )
            return func

        return inner

    @overload
    def receiver(
        self, response_type: ChunkResponseType
    ) -> Callable[[ReceiverFunc[dict[str, Any]]], ReceiverFunc[dict[str, Any]]]: ...

    @overload
    def receiver(
        self, response_type: TextResponseType
    ) -> Callable[[ReceiverFunc[str]], ReceiverFunc[str]]: ...
    @overload
    def receiver(
        self, response_type: AudioResponseType
    ) -> Callable[[ReceiverFunc[AudioSegment]], ReceiverFunc[AudioSegment]]: ...

    def receiver(
        self, response_type: ResponseType
    ) -> Callable[[ReceiverFunc[_ResponseT]], ReceiverFunc[_ResponseT]]:
        def inner(func: ReceiverFunc[_ResponseT]) -> ReceiverFunc[_ResponseT]:
            self.receivers[response_type].append(func)
            return func

        return inner

    def function_call(
        self, tool: type[BaseModel] | Callable
    ) -> Callable[[FunctionCallHandlerFunc], FunctionCallHandlerFunc]:
        def inner(func: FunctionCallHandlerFunc) -> FunctionCallHandlerFunc:
            base_tool = _convert_to_base_tool(tool)
            self.function_call_handlers[base_tool._name()] = func
            return func

        return inner

    async def _create_response(
        self,
        conn: ClientConnection,
        tools: list[type[BaseModel] | Callable] | None = None,
    ) -> None:
        if tools is not None:
            tool_schemas = _get_tool_type_and_tool_schemas(tools)
            schemas = []
            for tool_name, tool_type_and_tool_schema in tool_schemas.items():
                self._temporary_tool_schemas[tool_name].append(
                    tool_type_and_tool_schema
                )
                schemas.append(tool_type_and_tool_schema[1])
            event = {"type": "response.create", "response": {"tools": schemas}}
        else:
            event = {"type": "response.create"}
        await self._send_event(conn, event)

    async def _create_audio_input_buffer_commit(self, conn: ClientConnection) -> None:
        await self._send_event(conn, {"type": "input_audio_buffer.commit"})

    async def _create_conversation_item_create(
        self, conn: ClientConnection, message: str
    ) -> None:
        await self._send_event(
            conn,
            {
                "type": "conversation.item.create",
                "item": {
                    "type": "message",
                    "role": "user",
                    "content": [{"type": "input_text", "text": message}],
                },
            },
        )

    async def _create_session_update(
        self,
        conn: ClientConnection,
        tools: list[type[BaseModel] | Callable] | None = None,
    ) -> None:
        await self._send_event(
            conn,
            {
                "type": "session.update",
                "session": {
                    **self._session,
                    "tools": [
                        tool_schema
                        for _, tool_schema in _get_tool_type_and_tool_schemas(
                            tools
                        ).values()
                    ],
                }
                if tools
                else self._session,
            },
        )

    @classmethod
    async def _send_event(cls, conn: ClientConnection, event: dict[str, Any]) -> None:
        await conn.send(json.dumps(event))

    async def _wait_for_responses(self, sender: Sender) -> None:
        if sender["wait_for_text_response"]:
            await self._wait_for_text_message()
        if sender["wait_for_audio_transcript_response"]:
            await self._wait_for_audio_transcript()

    def _set_wait_for_flags(self, sender: Sender) -> None:
        if sender["wait_for_audio_transcript_response"]:
            self._audio_transcript_received = False
        if sender["wait_for_text_response"]:
            self._text_message_received = False

    async def _sender_process_loop(
        self, conn: ClientConnection, sender: Sender
    ) -> None:
        while True:
            await self._wait_for_responses(sender)
            if inspect.isasyncgenfunction(sender["func"]):
                input_audio_buffer: bool = False
                if tools := sender.get("tools"):
                    await self._create_session_update(conn, tools)
                async for message in sender["func"](self.context):
                    if isinstance(message, tuple):
                        message, tools = message
                        if tools:
                            await self._create_conversation_item_create(conn, message)
                    match message:
                        case BytesIO() | bytes():
                            event = await async_audio_input_audio_buffer_append_event(
                                message
                            )
                            await self._send_event(conn, event)
                            input_audio_buffer = True
                        case str():
                            await self._create_conversation_item_create(conn, message)
                        case _:
                            continue
                if input_audio_buffer:
                    await self._create_audio_input_buffer_commit(conn)
                await self._create_response(conn)
            else:
                tools = None
                message = await sender["func"](self.context)

                if isinstance(message, tuple):
                    message, tools = message
                    await self._create_conversation_item_create(conn, message)
                    await self._create_response(conn, tools)
                match message:
                    case BytesIO() | bytes():
                        event = await async_audio_to_item_create_event(message)
                        await self._send_event(conn, event)
                    case str():
                        await self._create_conversation_item_create(conn, message)
                    case _:
                        continue
                await self._create_response(conn, tools or sender.get("tools"))
            self._set_wait_for_flags(sender)

    async def _process_sender(self, conn: ClientConnection) -> None:
        try:
            tasks = (
                asyncio.create_task(self._sender_process_loop(conn, s))
                for s in self.senders
            )
            await asyncio.gather(*tasks)
        finally:
            await conn.close()

    async def _process_specific_function_call_handler(
        self,
        conn: ClientConnection,
        call_id: CallId,
        tool_name: ToolName,
        tool: OpenAIRealtimeTool,
    ) -> None:
        function_call_handler = self.function_call_handlers[tool_name]
        function_call_output = await function_call_handler(tool, self.context)
        await self._send_event(
            conn,
            {
                "type": "conversation.item.create",
                "item": {
                    "type": "function_call_output",
                    "call_id": call_id,
                    "output": function_call_output,
                },
            },
        )
        await self._send_event(conn, {"type": "response.create"})

    async def _process_specific_receiver(
        self,
        message: dict[str, Any] | str | AudioSegment | OpenAIRealtimeTool,
        response_type: ResponseType,
    ) -> None:
        for receiver in self.receivers[response_type]:
            await receiver(message, self.context)

    async def _process_receiver(self, conn: ClientConnection) -> None:
        try:
            audio_chunks: defaultdict[ResponseId, bytes] = defaultdict(bytes)
            text_chunks: defaultdict[ResponseId, str] = defaultdict(str)
            audio_transcript_chunks: defaultdict[ResponseId, str] = defaultdict(str)
            function_call_arguments_chunks: defaultdict[
                ResponseId, defaultdict[CallId, str]
            ] = defaultdict(lambda: defaultdict(str))
            async for data in conn:
                message = json.loads(data)  # type: RawMessage
                match message["type"]:  # type:
                    case "response.done":
                        if message["response"]["status"] == "failed":
                            print(  # noqa: T201
                                f"Failed: {message['response']['status_details']['error']}"
                            )
                    case "session.created":
                        current_session = message["session"]
                        self._session = {
                            key: current_session[key] if value == NotSet else value
                            for key, value in self._session.items()
                        }
                        await self._create_session_update(conn)
                    case "response.audio.done":
                        audio = audio_chunk_to_audio_segment(
                            audio_chunks.pop(message["response_id"])
                        )
                        await self._process_specific_receiver(audio, "audio")
                    case "response.audio.delta":
                        audio_chunks[message["response_id"]] += base64.b64decode(
                            message["delta"]
                        )
                        await self._process_specific_receiver(message, "audio_chunk")
                    case "response.text.done":
                        await self._process_specific_receiver(
                            text_chunks.pop(message["response_id"]), "text"
                        )
                        self._text_message_received = True
                    case "response.text.delta":
                        text_chunks[message["response_id"]] += message["delta"]
                        await self._process_specific_receiver(message, "text_chunk")
                    case "response.audio_transcript.done":
                        await self._process_specific_receiver(
                            audio_transcript_chunks[message["response_id"]],
                            "audio_transcript",
                        )
                        self._audio_transcript_received = True
                    case "response.audio_transcript.delta":
                        audio_transcript_chunks[message["response_id"]] += message[
                            "delta"
                        ]
                        await self._process_specific_receiver(
                            message, "audio_transcript_chunk"
                        )
                    case "response.function_call_arguments.delta":
                        function_call_arguments_chunks[message["response_id"]][
                            message["call_id"]
                        ] += message["delta"]
                        await self._process_specific_receiver(message, "tool_chunk")
                    case "response.function_call_arguments.done":
                        tool_name, response_id, call_id = (
                            message["name"],
                            message["response_id"],
                            message["call_id"],
                        )

                        tool_type_and_tool_schemas = self._temporary_tool_schemas.pop(
                            tool_name, []
                        )
                        if tool_type_and_tool_schemas:
                            tool_type, tool_schema = tool_type_and_tool_schemas.pop(0)
                        else:
                            tool_type, tool_schema = self._tool_schemas[tool_name]
                        function_call_arguments = function_call_arguments_chunks[
                            response_id
                        ].pop(call_id)
                        if not function_call_arguments_chunks[response_id]:
                            del function_call_arguments_chunks[response_id]
                        tool = tool_type.from_tool_call(
                            FunctionCallArguments(
                                call_id=call_id,
                                arguments=function_call_arguments,
                            )
                        )
                        await self._process_specific_function_call_handler(
                            conn, call_id, tool_name, tool
                        )

                # Call the receivers for all messages
                for receiver in self.receivers[None]:
                    await receiver(message, self.context)
        except asyncio.CancelledError:
            pass

    async def run(self) -> None:
        if self.is_running():
            raise RuntimeError("Already running")

        headers = {
            "Authorization": f"Bearer {os.getenv('OPENAI_API_KEY')}",
            "OpenAI-Beta": "realtime=v1",
        }

        async for conn in connect(
            uri=f"wss://api.openai.com/v1/realtime?model={self._session['model']}",
            additional_headers=headers,
        ):
            try:
                receiver_task = asyncio.create_task(self._process_receiver(conn))
                sender_task = asyncio.create_task(self._process_sender(conn))
                await asyncio.gather(receiver_task, sender_task)
            except websockets.ConnectionClosed:
                continue

    def is_running(self) -> bool:
        return self._connection is not None



from __future__ import annotations

import re
from typing import ClassVar

import requests
from bs4 import BeautifulSoup
from pydantic import Field

from mirascope.tools.base import ConfigurableTool, _ConfigurableToolConfig


class ParseURLConfig(_ConfigurableToolConfig):
    """Configuration for URL content parsing"""

    parser: str = Field(
        "html.parser",
        description="parser to use for parsing HTML content in BeautifulSoup",
        examples=["html.parser", "lxml", "lxml-xml", "html5lib"],
    )
    timeout: int = Field(
        default=5,
        description="Timeout in seconds for URL request",
    )


class ParseURLContent(ConfigurableTool[ParseURLConfig]):
    """Tool for parsing and extracting main content from URLs.

    Fetches content from URL, removes unnecessary elements like scripts, styles, navigation, etc.,
    and returns clean text content from the webpage's main body.
    """

    __configurable_tool_config__ = ParseURLConfig()  # pyright: ignore [reportCallIssue]
    __prompt_usage_description__: ClassVar[str] = """
    - `ParseURLContent`: Returns the given URL's main content as clean text
        - Fetches the raw HTML content for the given URL
        - Removes all unwanted elements (scripts, styles, navigation, etc.)
        - Attempts to find the main content section
        - If no main content section is found, extracts all text content
        - Returns cleaned and formatted text content
        - If any errors occur during fetching or parsing, and error message is returned.
    """

    url: str = Field(..., description="URL to fetch and parse")

    def call(self) -> str:
        """Fetch and parse content from the URL.

        Returns:
            str: Cleaned text content from the URL if successful, error message if parsing fails
        """
        try:
            # Fetch content from URL
            response = requests.get(
                self.url,
                timeout=self._get_config().timeout,
            )
            response.raise_for_status()

            # Parse HTML content
            soup = BeautifulSoup(response.text, self._get_config().parser)

            # Remove unwanted tags
            unwanted_tags = ["script", "style", "nav", "header", "footer", "aside"]
            for tag in unwanted_tags:
                for element in soup.find_all(tag):
                    element.decompose()

            # Find main content section
            main_content = (
                soup.find("main")
                or soup.find("article")
                or soup.find("div", class_=re.compile("content|main"))
            )

            # Extract and clean text
            if main_content:
                text = main_content.get_text(separator="\n", strip=True)
            else:
                text = soup.get_text(separator="\n", strip=True)

            # Remove empty lines and format
            lines = (line.strip() for line in text.splitlines())
            content = "\n".join(line for line in lines if line)

            if not content:
                return "No content found on the page"
            return content

        except requests.RequestException as e:
            return f"Failed to fetch content from URL: {str(e)}"
        except Exception as e:  # pragma: no cover
            return f"{type(e).__name__}: Failed to parse content from URL"



"""Tests the `_utils.extract_tool_return` module."""

from typing import Annotated

from pydantic import BaseModel, RootModel

from mirascope.core.base._utils._extract_tool_return import extract_tool_return
from mirascope.core.base.from_call_args import FromCallArgs


def test_extract_tool_return() -> None:
    """Tests the `extract_tool_return` function."""

    class Book(BaseModel):
        title: str
        author: str

    book = extract_tool_return(
        Book,
        '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
        allow_partial=False,
        fields_from_call_args={},
    )
    assert isinstance(book, Book)
    book = extract_tool_return(
        Book, '{"title": "The Name', allow_partial=True, fields_from_call_args={}
    )
    assert isinstance(book, Book)
    assert book.title == "The Name"
    assert book.author is None


def test_extract_tool_return_base_type() -> None:
    """Tests the `extract_tool_return` function with a base type."""

    title = extract_tool_return(
        str,
        '{"value": "The Name of the Wind"}',
        allow_partial=False,
        fields_from_call_args={},
    )
    assert isinstance(title, str)
    assert title == "The Name of the Wind"
    title = extract_tool_return(
        str, '{"value": "The Name', allow_partial=True, fields_from_call_args={}
    )
    assert isinstance(title, str)
    assert title == "The Name"


def test_extract_tool_return_parse_obj_with_fields_from_call_args() -> None:
    """Tests the `extract_tool_return` function parsing obj and fields from call args."""

    class Book(BaseModel):
        title: Annotated[str, FromCallArgs()]
        author: str

    book = extract_tool_return(
        Book,
        '{"author": "Patrick Rothfuss"}',
        allow_partial=False,
        fields_from_call_args={"title": "The Name of the Wind"},
    )
    assert isinstance(book, Book)
    assert book.title == "The Name of the Wind"
    assert book.author == "Patrick Rothfuss"


def test_extract_tool_return_parse_array_with_fields_from_call_args() -> None:
    """Tests the `extract_tool_return` function parsing array and fields from call args."""

    class Book(BaseModel):
        title: Annotated[str, FromCallArgs()]
        author: str

    # Nested models are not supported in this
    class ListModel(RootModel):
        root: list[Book]

    list_model = extract_tool_return(
        ListModel,
        '[{"author": "Patrick Rothfuss", "title": "The Name of the Wind"}]',
        allow_partial=False,
        fields_from_call_args={"title": "Dummy Title"},
    )
    assert isinstance(list_model, ListModel)
    assert isinstance(list_model.root, list)
    assert len(list_model.root) == 1
    assert list_model.root[0].title == "The Name of the Wind"



"""The `BedrockTool` class for easy tool usage with Bedrock LLM calls.

usage docs: learn/tools.md
"""

from __future__ import annotations

from mypy_boto3_bedrock_runtime.type_defs import (
    ToolSpecificationTypeDef,
    ToolTypeDef,
)
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool, GenerateJsonSchemaNoTitles, ToolConfig
from ._types import ToolUseBlockContentTypeDef


class GenerateBedrockStrictToolJsonSchema(GenerateJsonSchemaNoTitles):
    pass


class BedrockToolConfig(ToolConfig, total=False):
    """A tool configuration for Bedrock-specific features."""

    pass


class BedrockTool(BaseTool):
    """A class for defining tools for Bedrock LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.bedrock import bedrock_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @bedrock_call("anthropic.claude-3-haiku-20240307-v1:0", tools=[format_book])
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `BedrockTool` instance
        print(tool.call())
    ```
    """

    __provider__ = "bedrock"
    __tool_config_type__ = BedrockToolConfig

    tool_call: SkipJsonSchema[ToolUseBlockContentTypeDef]

    @property
    def id(self) -> str | None:
        return self.tool_call["toolUse"]["toolUseId"]

    @classmethod
    def tool_schema(cls) -> ToolTypeDef:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.bedrock import BedrockTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = BedrockTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Bedrock-specific tool schema
        ```
        """
        schema_generator = GenerateJsonSchemaNoTitles
        return ToolTypeDef(
            toolSpec=ToolSpecificationTypeDef(
                name=cls._name(),
                description=cls._description(),
                inputSchema={
                    "json": cls.model_json_schema(schema_generator=schema_generator)
                },
            )
        )

    @classmethod
    def from_tool_call(cls, tool_call: ToolUseBlockContentTypeDef) -> BedrockTool:
        """Constructs an `BedrockTool` instance from a `tool_call`.

        Args:
            tool_call: The Bedrock tool call from which to construct this tool instance.
        """
        return cls.model_validate(
            {"tool_call": tool_call, **tool_call["toolUse"]["input"]}
        )



"""Tests the `google.call_response_chunk` module."""

from google.genai.types import (
    Candidate,
    Content,
    GenerateContentResponse,
    Part,
)
from google.genai.types import (
    FinishReason as GoogleFinishReason,
)

from mirascope.core.base.types import CostMetadata
from mirascope.core.google.call_response_chunk import GoogleCallResponseChunk


def test_google_call_response_chunk() -> None:
    """Tests the `GoogleCallResponseChunk` class."""
    chunk = GenerateContentResponse(
        candidates=[
            Candidate(
                finish_reason=GoogleFinishReason.STOP,
                content=Content(
                    parts=[Part(text="The author is Patrick Rothfuss")],
                    role="model",
                ),
            )
        ]
    )

    call_response_chunk = GoogleCallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == "The author is Patrick Rothfuss"
    assert call_response_chunk.finish_reasons == [GoogleFinishReason.STOP]
    assert call_response_chunk.model is None
    assert call_response_chunk.id is None
    assert call_response_chunk.usage is None
    assert call_response_chunk.input_tokens is None
    assert call_response_chunk.output_tokens is None
    assert call_response_chunk.common_finish_reasons == ["stop"]
    assert call_response_chunk.cost_metadata == CostMetadata()

    call_response_chunk_no_candidates = GoogleCallResponseChunk(
        chunk=GenerateContentResponse(candidates=[])
    )
    assert call_response_chunk_no_candidates.content == ""



from __future__ import annotations

from collections.abc import Sequence
from typing import Any
from wave import Wave_read

from ._utils._convert_messages_to_message_params import (
    Image,
    convert_message_content_to_message_param_content,
)
from .message_param import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    DocumentPart,
    DocumentURLPart,
    ImagePart,
    ImageURLPart,
    TextPart,
)
from .types import AudioSegment


class Messages:
    Type = (
        str
        | Sequence[
            str
            | TextPart
            | CacheControlPart
            | ImagePart
            | ImageURLPart
            | Image.Image
            | AudioPart
            | AudioURLPart
            | AudioSegment
            | Wave_read
            | DocumentPart
            | DocumentURLPart
        ]
        | list[BaseMessageParam | Any]
        | BaseMessageParam
    )

    @classmethod
    def System(
        cls,
        content: str
        | Sequence[
            str
            | TextPart
            | CacheControlPart
            | ImagePart
            | ImageURLPart
            | Image.Image
            | AudioPart
            | AudioURLPart
            | AudioSegment
            | Wave_read
            | DocumentPart
            | DocumentURLPart
        ],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="system",
            content=convert_message_content_to_message_param_content(content),
        )

    @classmethod
    def User(
        cls,
        content: str
        | Sequence[
            str
            | TextPart
            | CacheControlPart
            | ImagePart
            | ImageURLPart
            | Image.Image
            | AudioPart
            | AudioURLPart
            | AudioSegment
            | Wave_read
            | DocumentPart
            | DocumentURLPart
        ],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="user",
            content=convert_message_content_to_message_param_content(content),
        )

    @classmethod
    def Assistant(
        cls,
        content: str
        | Sequence[
            str
            | TextPart
            | CacheControlPart
            | ImagePart
            | ImageURLPart
            | Image.Image
            | AudioPart
            | AudioURLPart
            | AudioSegment
            | Wave_read
            | DocumentPart
            | DocumentURLPart
        ],
    ) -> BaseMessageParam:
        return BaseMessageParam(
            role="assistant",
            content=convert_message_content_to_message_param_content(content),
        )



"""This module contains the setup_call function for Groq tools."""

import inspect
from collections.abc import Awaitable, Callable
from typing import Any, cast, overload

from groq import AsyncGroq, Groq
from groq.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
)
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import AsyncCreateFn, CreateFn, get_async_create_fn, get_create_fn
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import GroqCallKwargs
from ..call_params import GroqCallParams
from ..dynamic_config import AsyncGroqDynamicConfig, GroqDynamicConfig
from ..tool import GroqTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: AsyncGroq | None,
    fn: Callable[..., Awaitable[AsyncGroqDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncGroqDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GroqCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[GroqTool]] | None,
    GroqCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: Groq | None,
    fn: Callable[..., GroqDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: GroqDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GroqCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[GroqTool]] | None,
    GroqCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: Groq | AsyncGroq | None,
    fn: Callable[..., GroqDynamicConfig | Awaitable[AsyncGroqDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GroqDynamicConfig | AsyncGroqDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GroqCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk]
    | AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[GroqTool]] | None,
    GroqCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        GroqTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(GroqCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | ChatCompletionMessageParam], messages)
    messages = convert_message_params(messages)
    if json_mode:
        if not tools:
            call_kwargs["response_format"] = {"type": "json_object"}
        json_mode_content = _utils.json_mode_content(response_model)
        if messages[-1]["role"] == "user":
            if isinstance(messages[-1]["content"], str):
                messages[-1]["content"] += json_mode_content
            else:
                messages[-1]["content"] = list(messages[-1]["content"]) + [
                    {"type": "text", "text": json_mode_content.strip()}
                ]
        else:
            messages.append({"role": "user", "content": json_mode_content.strip()})
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = {
            "type": "function",
            "function": {"name": tool_types[0]._name()},
        }
    call_kwargs |= {"model": model, "messages": messages}
    if client is None:
        client = AsyncGroq() if inspect.iscoroutinefunction(fn) else Groq()

    create = (
        get_async_create_fn(client.chat.completions.create)
        if isinstance(client, AsyncGroq)
        else get_create_fn(client.chat.completions.create)
    )

    return create, prompt_template, messages, tool_types, call_kwargs



"""The CallResponse class for the LLM provider."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import Any

from pydantic import computed_field

from ..core import BaseDynamicConfig
from ..core.base import (
    BaseCallParams,
    BaseCallResponse,
    BaseMessageParam,
    BaseTool,
    Usage,
    transform_tool_outputs,
)
from ..core.base._utils import BaseMessageParamConverter
from ..core.base.message_param import ToolResultPart
from ..core.base.types import FinishReason
from ._response_metaclass import _ResponseMetaclass
from .tool import Tool


class CallResponse(
    BaseCallResponse[
        Any,
        BaseTool,
        Any,
        BaseDynamicConfig[Any, Any, Any],
        BaseMessageParam,
        BaseCallParams,
        BaseMessageParam,
        BaseMessageParamConverter,
    ],
    metaclass=_ResponseMetaclass,
):
    """
    A provider-agnostic CallResponse class.

    We rely on _response having `common_` methods or properties for normalization.
    """

    _response: BaseCallResponse[Any, BaseTool, Any, Any, Any, Any, Any, Any]

    def __init__(
        self,
        response: BaseCallResponse[Any, BaseTool, Any, Any, Any, Any, Any, Any],
    ) -> None:
        super().__init__(
            **{
                field: getattr(response, field)
                for field in response.model_fields
                if field != "user_message_param"
            }
        )
        object.__setattr__(self, "_response", response)
        object.__setattr__(
            self, "user_message_param", response.common_user_message_param
        )
        object.__setattr__(self, "messages", response.common_messages)

    def __getattribute__(self, name: str) -> Any:  # noqa: ANN401
        special_names = {
            "_response",
            "common_messages",
            "finish_reasons",
            "usage",
            "message_param",
            "user_message_param",
            "messages",
            "tools",
            "tool",
            "tool_message_params",
            "__dict__",
            "__class__",
            "model_fields",
            "__annotations__",
            "__pydantic_validator__",
            "__pydantic_fields_set__",
            "__pydantic_extra__",
            "__pydantic_private__",
            "__class_getitem__",
            "__repr__",
            "__str__",
            "_properties",
        } | set(object.__getattribute__(self, "_properties"))

        if name in special_names:
            return object.__getattribute__(self, name)

        try:
            response = object.__getattribute__(self, "_response")
            return getattr(response, name)
        except AttributeError:
            return object.__getattribute__(self, name)

    def __str__(self) -> str:
        return str(self._response)

    @computed_field
    @property
    def finish_reasons(self) -> list[FinishReason] | None:  # pyright: ignore [reportIncompatibleMethodOverride]
        return self._response.common_finish_reasons

    @property
    def usage(self) -> Usage | None:
        """Returns the usage of the chat completion."""
        return self._response.common_usage

    @computed_field
    @cached_property
    def message_param(self) -> BaseMessageParam:
        return self._response.common_message_param  # pyright: ignore [reportReturnType]

    @computed_field
    @cached_property
    def common_messages(self) -> list[BaseMessageParam]:  # pyright: ignore [reportIncompatibleMethodOverride]
        return self._response.common_messages

    @cached_property
    def tools(self) -> list[Tool] | None:  # pyright: ignore [reportIncompatibleVariableOverride]
        return self._response.common_tools

    @cached_property
    def tool(self) -> Tool | None:  # pyright: ignore [reportIncompatibleVariableOverride]
        tools = self._response.common_tools
        if tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[BaseTool, str]]
    ) -> list[BaseMessageParam]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.
        """
        return [
            BaseMessageParam(
                role="tool",
                content=[
                    ToolResultPart(
                        type="tool_result",
                        name=tool._name(),
                        content=output,
                        id=tool.id,
                    )
                ],
            )
            for tool, output in tools_and_outputs
        ]



"""This module contains the setup_call function for the Anthropic API."""

import inspect
from collections.abc import Awaitable, Callable
from typing import Any, cast, overload

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)
from anthropic.types import Message, MessageParam, MessageStreamEvent
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import AsyncCreateFn, CreateFn
from ...base.stream_config import StreamConfig
from .._call_kwargs import AnthropicCallKwargs
from ..call_params import AnthropicCallParams
from ..dynamic_config import AnthropicDynamicConfig, AsyncAnthropicDynamicConfig
from ..tool import AnthropicTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: AsyncAnthropic | AsyncAnthropicBedrock | AsyncAnthropicVertex | None,
    fn: Callable[..., Awaitable[AsyncAnthropicDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncAnthropicDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AnthropicCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[Message, MessageStreamEvent],
    str | None,
    list[MessageParam],
    list[type[AnthropicTool]] | None,
    AnthropicCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: Anthropic | AnthropicBedrock | AnthropicVertex | None,
    fn: Callable[..., AnthropicDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: AnthropicDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AnthropicCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[Message, MessageStreamEvent],
    str | None,
    list[MessageParam],
    list[type[AnthropicTool]] | None,
    AnthropicCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: Anthropic
    | AsyncAnthropic
    | AnthropicBedrock
    | AsyncAnthropicBedrock
    | AnthropicVertex
    | AsyncAnthropicVertex
    | None,
    fn: Callable[..., AnthropicDynamicConfig | Awaitable[AsyncAnthropicDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncAnthropicDynamicConfig | AnthropicDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AnthropicCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    Callable[..., Message | Awaitable[Message]],
    str | None,
    list[MessageParam],
    list[type[AnthropicTool]] | None,
    AnthropicCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        AnthropicTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(AnthropicCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | MessageParam], messages)
    messages = convert_message_params(messages)

    if messages[0]["role"] == "system":
        call_kwargs["system"] = messages.pop(0)["content"]  # pyright: ignore [reportGeneralTypeIssues]

    if json_mode:
        json_mode_content = _utils.json_mode_content(response_model)
        if isinstance(messages[-1]["content"], str):
            messages[-1]["content"] += json_mode_content
        else:
            messages[-1]["content"] = list(messages[-1]["content"]) + [
                {"type": "text", "text": json_mode_content}
            ]
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = {"type": "tool", "name": tool_types[0]._name()}
    call_kwargs |= {
        "model": model,
        "messages": messages,
        "max_tokens": call_kwargs["max_tokens"],
    }

    if client is None:
        client = AsyncAnthropic() if inspect.iscoroutinefunction(fn) else Anthropic()
    create = client.messages.create
    return create, prompt_template, messages, tool_types, call_kwargs



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from google.genai.types import (
    GenerateContentConfigOrDict,
)
from typing_extensions import NotRequired

from ..base import BaseCallParams


class GoogleCallParams(BaseCallParams):
    """The parameters to use when calling the Google API.

    [Google API Reference](https://ai.google.dev/google-api/docs/text-generation?lang=python)

    Attributes:
        config: ...
    """

    config: NotRequired[GenerateContentConfigOrDict]



"""The Mirascope LiteLLM Module."""

from typing import TypeAlias

from ..openai import OpenAIMessageParam
from ._call import litellm_call
from ._call import litellm_call as call
from .call_params import LiteLLMCallParams
from .call_response import LiteLLMCallResponse
from .call_response_chunk import LiteLLMCallResponseChunk
from .dynamic_config import AsyncLiteLLMDynamicConfig, LiteLLMDynamicConfig
from .stream import LiteLLMStream
from .tool import LiteLLMTool

LiteLLMMessageParam: TypeAlias = OpenAIMessageParam

__all__ = [
    "AsyncLiteLLMDynamicConfig",
    "LiteLLMCallParams",
    "LiteLLMCallResponse",
    "LiteLLMCallResponseChunk",
    "LiteLLMDynamicConfig",
    "LiteLLMMessageParam",
    "LiteLLMStream",
    "LiteLLMTool",
    "call",
    "litellm_call",
]



import pytest
from google.generativeai import protos

from mirascope.core import BaseMessageParam
from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    DocumentPart,
    ImagePart,
    ImageURLPart,
)
from mirascope.core.base.message_param import ToolCallPart, ToolResultPart
from mirascope.core.gemini._utils._message_param_converter import (
    GeminiMessageParamConverter,
    _to_audio_part,
    _to_document_part,
    _to_image_part,
)


def test_gemini_convert_parts_text_only():
    """
    If parts contain only text, the converter should produce
    a BaseMessageParam with either a single text string or a list[TextPart].
    """
    part = protos.Part(text="hello world")
    results = GeminiMessageParamConverter.from_provider(
        [
            {
                "role": "assistant",
                "parts": [part],
            }
        ]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"

    assert result.content == "hello world"


def test_gemini_convert_parts_image():
    """
    If inline_data has an image mime type, produce an ImagePart.
    """
    part = protos.Part(
        inline_data=protos.Blob(mime_type="image/png", data=b"\x89PNG\r\n\x1a\n")
    )
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    assert isinstance(result.content[0], ImagePart)
    assert result.content[0].media_type == "image/png"
    assert result.content[0].image == b"\x89PNG\r\n\x1a\n"


def test_gemini_convert_parts_audio():
    part = protos.Part(inline_data=protos.Blob(mime_type="audio/mp3", data=b"audio"))
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    assert isinstance(result.content[0], AudioPart)
    assert result.content[0].media_type == "audio/mp3"
    assert result.content[0].audio == b"audio"


def test_gemini_convert_parts_document():
    """
    If file_data is a PDF, produce a DocumentPart.
    """
    part = protos.Part(
        file_data=protos.FileData(mime_type="application/pdf", file_uri=b"%PDF-1.4...")
    )
    with pytest.raises(
        ValueError,
        match="Unsupported file_data mime type: application/pdf. Cannot convert to BaseMessageParam.",
    ):
        GeminiMessageParamConverter.from_provider(
            [{"role": "assistant", "parts": [part]}]
        )


def test_gemini_convert_parts_unsupported_image():
    """
    If inline_data has an unsupported image type (e.g. image/tiff),
    the converter should raise ValueError.
    """
    part = protos.Part(inline_data=protos.Blob(mime_type="image/tiff", data=b"fake"))
    with pytest.raises(
        ValueError,
        match="Unsupported inline_data mime type: image/tiff. Cannot convert to BaseMessageParam.",
    ):
        GeminiMessageParamConverter.from_provider(
            [{"role": "assistant", "parts": [part]}]
        )


def test_gemini_convert_function_call():
    """
    If the part has a function_call, the converter returns a new BaseMessageParam
    with a ToolCallPart inside.
    """
    part = protos.Part(
        function_call=protos.FunctionCall(name="some_tool", args={"param1": "value1"})
    )
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    part_call = result.content[0]
    assert isinstance(part_call, ToolCallPart)
    assert part_call.name == "some_tool"
    assert part_call.args == {"param1": "value1"}


def test_gemini_convert_no_supported_content():
    """
    If none of text, inline_data, file_data, function_call, or function_response
    is set, the converter should raise ValueError.
    """
    part = protos.Part()
    # No text, no inline_data, no file_data, no function_call, no function_response
    with pytest.raises(
        ValueError,
        match="Part does not contain any supported content \\(text, image, or document\\).",
    ):
        GeminiMessageParamConverter.from_provider(
            [{"role": "assistant", "parts": [part]}]
        )


def test_to_image_part_unsupported_image():
    with pytest.raises(
        ValueError,
        match="Unsupported image media type: application/json. Expected one of: image/jpeg, image/png, image/gif, image/webp.",
    ):
        _to_image_part(mime_type="application/json", data=b"{}")


def test_to_document_part_unsupported_document():
    with pytest.raises(
        ValueError,
        match="Unsupported document media type: application/json. Only application/pdf is supported.",
    ):
        _to_document_part(mime_type="application/json", data=b"{}")


def test_to_audio_part_supported():
    """Test _to_audio_part with a supported audio mime type."""
    result = _to_audio_part(mime_type="audio/mp3", data=b"audio")
    assert isinstance(result, AudioPart)
    assert result.media_type == "audio/mp3"
    assert result.audio == b"audio"


def test_to_audio_part_unsupported():
    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: application/json. Expected one of: audio/wav, audio/mp3, audio/aiff, audio/aac, audio/ogg, audio/flac.",
    ):
        _to_audio_part(mime_type="application/json", data=b"{}")


def test_to_provider():
    results = GeminiMessageParamConverter.to_provider(
        [BaseMessageParam(role="assistant", content="Hello")]
    )
    assert results == [{"parts": ["Hello"], "role": "model"}]


def test_inline_data_application_pdf():
    part = protos.Part(
        inline_data=protos.Blob(mime_type="application/pdf", data=b"%PDF-1.4...")
    )
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    assert isinstance(result.content[0], DocumentPart)
    assert result.content[0].media_type == "application/pdf"
    assert result.content[0].document == b"%PDF-1.4..."


def test_function_response():
    part = protos.Part(
        function_response=protos.FunctionResponse(
            name="some_tool", response={"result": "value"}
        )
    )
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    assert results == [
        BaseMessageParam(
            role="assistant",
            content=[
                ToolResultPart(
                    type="tool_result",
                    name="some_tool",
                    content="value",
                    id=None,
                    is_error=False,
                )
            ],
        )
    ]


def test_protos_function_response():
    part = protos.FunctionResponse(name="some_tool", response={"result": "value"})

    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    assert results == [
        BaseMessageParam(
            role="assistant",
            content=[
                ToolResultPart(
                    type="tool_result",
                    name="some_tool",
                    content="value",
                    id=None,
                    is_error=False,
                )
            ],
        )
    ]


def test_protos_function_call():
    part = protos.FunctionCall(name="some_tool", args={"param1": "value1"})

    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    part_call = result.content[0]
    assert isinstance(part_call, ToolCallPart)
    assert part_call.name == "some_tool"
    assert part_call.args == {"param1": "value1"}


def test_gemini_convert_parts_file_data_image():
    """
    When a part has file_data with an image mime type, it should produce an ImageURLPart.
    """
    file_data = protos.FileData(
        mime_type="image/jpeg", file_uri="http://example.com/img"
    )
    part = protos.Part(file_data=file_data)
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    assert isinstance(result.content[0], ImageURLPart)
    assert result.content[0].url == "http://example.com/img"


def test_gemini_convert_parts_file_data_audio():
    """
    When a part has file_data with an audio mime type, it should produce an AudioURLPart.
    """
    file_data = protos.FileData(
        mime_type="audio/mp3", file_uri="http://example.com/audio"
    )
    part = protos.Part(file_data=file_data)
    results = GeminiMessageParamConverter.from_provider(
        [{"role": "assistant", "parts": [part]}]
    )
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    assert isinstance(result.content[0], AudioURLPart)
    assert result.content[0].url == "http://example.com/audio"


def test_gemini_convert_parts_file_data_unsupported():
    """
    When a part has file_data with an unsupported mime type, the converter should raise ValueError.
    """
    file_data = protos.FileData(
        mime_type="application/zip", file_uri="http://example.com/zip"
    )
    part = protos.Part(file_data=file_data)
    with pytest.raises(
        ValueError,
        match="Unsupported file_data mime type: application/zip. Cannot convert to BaseMessageParam.",
    ):
        GeminiMessageParamConverter.from_provider(
            [{"role": "assistant", "parts": [part]}]
        )



"""
--------------------------------------------------------------------------------
Source: https://github.com/pydantic/pydantic/issues/6381#issuecomment-1831607091
By: silviumarcu

This code is used in accordance with the repository's license, and this reference
serves as an acknowledgment of the original author's contribution to this project.
--------------------------------------------------------------------------------
"""

from copy import deepcopy
from typing import TypeVar, get_args, get_origin

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

Model = TypeVar("Model", bound=BaseModel)


def _process_annotation(annotation: type) -> type:
    """Recursively process type annotations to make them optional."""
    # Get the origin type (e.g., list from list[str])
    origin = get_origin(annotation)

    if origin is None:
        # If it's a BaseModel, make it partial
        if isinstance(annotation, type) and issubclass(annotation, BaseModel):
            return partial(annotation) | None  # pyright: ignore [reportReturnType]
        # For simple types, just make them optional
        return annotation | None  # pyright: ignore [reportReturnType]

    # Get the type arguments (e.g., str from list[str])
    args = get_args(annotation)

    # Recursively process each type argument
    processed_args = tuple(
        _process_annotation(arg) if isinstance(arg, type) else arg for arg in args
    )

    # Reconstruct the type with processed arguments
    return (
        origin[processed_args[0] if len(processed_args) == 1 else processed_args] | None
    )  # pyright: ignore [reportReturnType]


def partial(
    wrapped_class: type[Model], preserve_fields: set[str] | None = None
) -> type[Model]:
    """Generate a new class with all attributes optionals.

    This decorator will wrap a class inheriting from BaseModel and will recursively
    convert all its attributes and its children's attributes to optionals, including
    handling generic type hints like list[Model].

    Example:
    ```python
    @partial
    class User(BaseModel):
        name: str
        friends: list[Friend]  # Will become list[PartialFriend] | None

    user = User()  # All fields optional
    ```
    """
    if preserve_fields is None:
        preserve_fields = set()

    def _make_field_optional(
        field: FieldInfo,
    ) -> tuple[object, FieldInfo]:
        tmp_field = deepcopy(field)

        # Process the field's annotation
        tmp_field.annotation = _process_annotation(field.annotation)  # pyright: ignore [reportArgumentType]

        # Set default value
        tmp_field.default = None

        return tmp_field.annotation, tmp_field

    return create_model(
        f"Partial{wrapped_class.__name__}",
        __base__=wrapped_class,
        __module__=wrapped_class.__module__,
        __doc__=wrapped_class.__doc__,
        __config__=None,
        __validators__=None,
        __cls_kwargs__=None,
        **{
            field_name: (field_info.annotation, field_info)
            if field_name in preserve_fields
            else _make_field_optional(field_info)
            for field_name, field_info in wrapped_class.model_fields.items()
        },
    )



"""Tests the `_utils.get_audio_type` module."""

import pytest

from mirascope.core.base._utils._get_audio_type import get_audio_type

WAV_DATA = b"RIFF$\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac\x00\x00\x88X\x01\x00\x02\x00\x10\x00data\x00\x00\x00\x00"

MP3_DATA = b"ID3\x03\x00\x00\x00\x00\x00\x1f"

AIFF_DATA = b"FORM\x00\x00\x00\x00AIFFCOMM"

AAC_DATA = b"\xff\xf1"

OGG_DATA = b"OggS\x00\x02"

FLAC_DATA = b"fLaC\x00\x00\x00\x22\x10\x00\x00\x00\x01"


def test_get_audio_type() -> None:
    """Test the get_audio_type function."""
    assert get_audio_type(WAV_DATA) == "wav"
    assert get_audio_type(MP3_DATA) == "mp3"
    assert get_audio_type(AIFF_DATA) == "aiff"
    assert get_audio_type(AAC_DATA) == "aac"
    assert get_audio_type(OGG_DATA) == "ogg"
    assert get_audio_type(FLAC_DATA) == "flac"
    with pytest.raises(ValueError, match="Unsupported audio type"):
        get_audio_type(b"invalid")



"""Utility for converting `BaseMessageParam` to `ContentsType`"""

import asyncio
import base64
import io
from concurrent.futures import ThreadPoolExecutor

from google.genai import Client
from google.genai.types import (
    BlobDict,
    ContentDict,
    FileDataDict,
    FunctionCallDict,
    FunctionResponseDict,
    PartDict,
)

from ...base import BaseMessageParam
from ...base._utils import get_audio_type, get_document_type, get_image_type
from ...base._utils._parse_content_template import _load_media
from ._validate_media_type import (
    _check_audio_media_type,
    _check_document_media_type,
    _check_image_media_type,
)


def _over_file_size_limit(size: int) -> bool:
    """Check if the total file size exceeds the limit (10mb).

    Google limit is 20MB but base64 adds 33% to the size.
    """
    return size > 10 * 1024 * 1024  # 10MB


async def _convert_message_params_async(
    message_params: list[BaseMessageParam | ContentDict], client: Client
) -> list[ContentDict]:
    converted_message_params = []
    total_payload_size = 0
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif (role := message_param.role) == "system":
            if not isinstance(message_param.content, str):
                raise ValueError(
                    "System message content must be a single text string."
                )  # pragma: no cover
            converted_message_params += [
                {
                    "role": "system",
                    "parts": [PartDict(text=message_param.content)],
                }
            ]
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(
                {
                    "role": role if role == "user" else "model",
                    "parts": [PartDict(text=content)],
                }
            )
        else:
            converted_content = []
            must_upload: dict[int, BlobDict] = {}
            for index, part in enumerate(content):
                if part.type == "text":
                    converted_content.append(PartDict(text=part.text))
                elif part.type == "tool_call":
                    converted_content.append(
                        PartDict(
                            function_call=FunctionCallDict(
                                name=part.name, args=part.args, id=part.id
                            )
                        )
                    )
                elif part.type == "tool_result":
                    converted_content.append(
                        PartDict(
                            function_response=FunctionResponseDict(
                                name=part.name,
                                response={"content": part.content},
                                id=part.id,
                            )
                        )
                    )
                elif part.type == "image":
                    _check_image_media_type(part.media_type)
                    blob_dict = BlobDict(data=part.image, mime_type=part.media_type)
                    converted_content.append(PartDict(inline_data=blob_dict))
                    image_size = len(part.image)
                    total_payload_size += image_size
                    if _over_file_size_limit(total_payload_size):
                        must_upload[index] = blob_dict
                        total_payload_size -= image_size
                elif part.type == "image_url":
                    if (
                        client.vertexai
                        or not part.url.startswith(("https://", "http://"))
                        or "generativelanguage.googleapis.com" in part.url
                    ):
                        converted_content.append(
                            PartDict(
                                file_data=FileDataDict(
                                    file_uri=part.url, mime_type=None
                                )
                            )
                        )
                    else:
                        downloaded_image = _load_media(part.url)
                        media_type = f"image/{get_image_type(downloaded_image)}"
                        _check_image_media_type(media_type)
                        blob_dict = BlobDict(
                            data=downloaded_image, mime_type=media_type
                        )
                        converted_content.append(PartDict(inline_data=blob_dict))
                        image_size = len(downloaded_image)
                        total_payload_size += image_size
                        if _over_file_size_limit(total_payload_size):
                            must_upload[index] = blob_dict
                            total_payload_size -= image_size
                elif part.type == "audio":
                    _check_audio_media_type(part.media_type)
                    audio_data = (
                        part.audio
                        if isinstance(part.audio, bytes)
                        else base64.b64decode(part.audio)
                    )
                    blob_dict = BlobDict(data=audio_data, mime_type=part.media_type)
                    converted_content.append(PartDict(inline_data=blob_dict))
                    audio_size = len(audio_data)
                    total_payload_size += audio_size
                    if _over_file_size_limit(total_payload_size):
                        must_upload[index] = blob_dict
                        total_payload_size -= audio_size
                elif part.type == "audio_url":
                    if (
                        client.vertexai
                        or not part.url.startswith(("https://", "http://"))
                        or "generativelanguage.googleapis.com" in part.url
                    ):
                        converted_content.append(
                            PartDict(
                                file_data=FileDataDict(
                                    file_uri=part.url, mime_type=None
                                )
                            )
                        )
                    else:
                        downloaded_audio = _load_media(part.url)
                        media_type = f"audio/{get_audio_type(downloaded_audio)}"
                        _check_audio_media_type(media_type)
                        blob_dict = BlobDict(
                            data=downloaded_audio, mime_type=media_type
                        )
                        converted_content.append(PartDict(inline_data=blob_dict))
                        audio_size = len(downloaded_audio)
                        total_payload_size += audio_size
                        if _over_file_size_limit(total_payload_size):
                            must_upload[index] = blob_dict
                            total_payload_size -= audio_size
                elif part.type == "document":
                    _check_document_media_type(part.media_type)
                    blob_dict = BlobDict(data=part.document, mime_type=part.media_type)
                    converted_content.append(PartDict(inline_data=blob_dict))
                    document_size = len(part.document)
                    total_payload_size += document_size
                    if _over_file_size_limit(total_payload_size):
                        must_upload[index] = blob_dict
                        total_payload_size -= document_size
                elif part.type == "document_url":
                    if (
                        client.vertexai
                        or not part.url.startswith(("https://", "http://"))
                        or "generativelanguage.googleapis.com" in part.url
                    ):
                        converted_content.append(
                            PartDict(
                                file_data=FileDataDict(
                                    file_uri=part.url, mime_type=None
                                )
                            )
                        )
                    else:
                        media_type = None
                        try:
                            downloaded_document = _load_media(part.url)
                            document_types = {
                                "pdf": "application/pdf",
                                "html": "text/html",
                                "xml": "text/xml",
                                "rtf": "text/rtf",
                            }

                            try:
                                document_type_ext = get_document_type(
                                    downloaded_document
                                )
                                media_type = document_types.get(document_type_ext)
                            except ValueError:
                                try:
                                    downloaded_document.decode("utf-8")
                                    media_type = "text/plain"
                                except UnicodeDecodeError:
                                    pass

                            if media_type is None:
                                raise ValueError(
                                    f"Unsupported document format detected for URL: {part.url}. "
                                    "Google API only supports the following document formats: "
                                    "PDF, JavaScript, Python, TXT, HTML, CSS, Markdown, CSV, XML and RTF. "
                                    "Please provide a document in one of these supported formats."
                                )

                            _check_document_media_type(media_type)
                            blob_dict = BlobDict(
                                data=downloaded_document, mime_type=media_type
                            )
                            converted_content.append(PartDict(inline_data=blob_dict))
                            document_size = len(downloaded_document)
                            total_payload_size += document_size
                            if _over_file_size_limit(total_payload_size):
                                must_upload[index] = blob_dict
                                total_payload_size -= document_size
                        except ValueError as e:
                            raise ValueError(
                                f"Failed to process document from URL: {part.url}. "
                                f"Error details: {str(e)}. "
                                "Please ensure the URL is accessible and points to a supported document format."
                            )
                else:
                    raise ValueError(
                        "Google currently only supports text, tool_call, tool_result, image, audio, and document parts. "
                        f"Part provided: {part.type}"
                    )

            if must_upload:
                indices, blob_dicts = zip(*must_upload.items(), strict=True)
                upload_tasks = [
                    client.aio.files.upload(
                        file=io.BytesIO(blob_dict["data"]),
                        config={"mime_type": blob_dict.get("mime_type", None)},
                    )
                    for blob_dict in blob_dicts
                ]
                file_refs = await asyncio.gather(*upload_tasks)
                for index, file_ref in zip(indices, file_refs, strict=True):
                    converted_content[index] = PartDict(
                        file_data=FileDataDict(
                            file_uri=file_ref.uri, mime_type=file_ref.mime_type
                        )
                    )

            converted_message_params.append(
                {
                    "role": role if role == "user" else "model",
                    "parts": converted_content,
                }
            )
    return converted_message_params


def convert_message_params(
    message_params: list[BaseMessageParam | ContentDict], client: Client
) -> list[ContentDict]:
    """Convert message params to Google's ContentDict format.

    If called from sync context, uses asyncio.run().
    If called from async context, uses the current event loop.
    """
    try:
        asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(
                asyncio.run, _convert_message_params_async(message_params, client)
            )
            return future.result()
    except RuntimeError:
        ...
    return asyncio.run(_convert_message_params_async(message_params, client))



"""Utility for determining the type of an document from its bytes."""


def get_document_type(document_data: bytes) -> str:
    if document_data.startswith(b"%PDF"):
        return "pdf"
    raise ValueError("Unsupported document type")



"""This module contains the setup_call function for Mistral tools."""

import os
from collections.abc import (
    Awaitable,
    Callable,
)
from typing import Any, cast, overload

from mistralai import Mistral
from mistralai.models import (
    AssistantMessage,
    ChatCompletionResponse,
    CompletionEvent,
    ResponseFormat,
    SystemMessage,
    TextChunk,
    ToolChoiceEnum,
    ToolMessage,
    UserMessage,
)
from pydantic import BaseModel

from ... import BaseMessageParam
from ...base import BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    fn_is_async,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import MistralCallKwargs
from ..call_params import MistralCallParams
from ..dynamic_config import MistralDynamicConfig
from ..tool import MistralTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: Mistral | None,
    fn: Callable[..., Awaitable[MistralDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: MistralDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: MistralCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[ChatCompletionResponse, CompletionEvent],
    str | None,
    list[AssistantMessage | SystemMessage | ToolMessage | UserMessage],
    list[type[MistralTool]] | None,
    MistralCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: Mistral | None,
    fn: Callable[..., MistralDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: MistralDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: MistralCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletionResponse, CompletionEvent],
    str | None,
    list[AssistantMessage | SystemMessage | ToolMessage | UserMessage],
    list[type[MistralTool]] | None,
    MistralCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: Mistral | None,
    fn: Callable[..., MistralDynamicConfig | Awaitable[MistralDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: MistralDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: MistralCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletionResponse, CompletionEvent]
    | AsyncCreateFn[ChatCompletionResponse, CompletionEvent],
    str | None,
    list[AssistantMessage | SystemMessage | ToolMessage | UserMessage],
    list[type[MistralTool]] | None,
    MistralCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        MistralTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(MistralCallKwargs, base_call_kwargs)
    messages = cast(
        list[
            BaseMessageParam
            | AssistantMessage
            | SystemMessage
            | ToolMessage
            | UserMessage
        ],
        messages,
    )
    messages = convert_message_params(messages)
    if json_mode:
        if not tools:
            call_kwargs["response_format"] = ResponseFormat(type="json_object")
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        if messages[-1].role == "user":
            if isinstance(messages[-1].content, list):
                messages[-1].content.append(TextChunk(text=json_mode_content))
            elif isinstance(messages[-1].content, str):
                messages[-1].content += json_mode_content
        else:
            messages.append(UserMessage(content=json_mode_content.strip()))
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = cast(ToolChoiceEnum, "any")
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = Mistral(api_key=os.environ["MISTRAL_API_KEY"])
    if fn_is_async(fn):
        create_or_stream = get_async_create_fn(
            client.chat.complete_async, client.chat.stream_async
        )
    else:
        create_or_stream = get_create_fn(client.chat.complete, client.chat.stream)
    return create_or_stream, prompt_template, messages, tool_types, call_kwargs



"""Tests the `groq.call_response` module."""

from groq.types.chat import ChatCompletion, ChatCompletionToolMessageParam
from groq.types.chat.chat_completion import Choice
from groq.types.chat.chat_completion_message import ChatCompletionMessage
from groq.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from groq.types.completion_usage import CompletionUsage

from mirascope.core import BaseMessageParam
from mirascope.core.base.types import ImageMetadata
from mirascope.core.groq.call_response import GroqCallResponse
from mirascope.core.groq.tool import GroqTool


def test_groq_call_response() -> None:
    """Tests the `GroqCallResponse` class."""
    choices = [
        Choice(
            finish_reason="stop",
            index=0,
            message=ChatCompletionMessage(content="content", role="assistant"),
        )
    ]
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    completion = ChatCompletion(
        id="id",
        choices=choices,
        created=0,
        model="llama3-70b-8192",
        object="chat.completion",
        usage=usage,
    )
    call_response = GroqCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "groq"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["stop"]
    assert call_response.model == "llama3-70b-8192"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 1.38e-6
    assert call_response.message_param == {
        "content": "content",
        "role": "assistant",
        "tool_calls": None,
    }
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant", content="content"
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = {"role": "user", "content": "content"}
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_groq_call_response_with_tools() -> None:
    """Tests the `GroqCallResponse` class with tools."""

    class FormatBook(GroqTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ChatCompletionMessageToolCall(
        id="id",
        function=Function(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    completion = ChatCompletion(
        id="id",
        choices=[
            Choice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content="content", role="assistant", tool_calls=[tool_call]
                ),
            )
        ],
        created=0,
        model="llama3-70b-8192",
        object="chat.completion",
    )
    call_response = GroqCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        ChatCompletionToolMessageParam(
            role="tool",
            content=output,
            tool_call_id=tool_call.id,
            name="FormatBook",  # type: ignore
        )
    ]


def test_cost_metadata_images() -> None:
    """Test the cost_metadata property to ensure image metadata is computed correctly."""
    choices = [
        Choice(
            finish_reason="stop",
            index=0,
            message=ChatCompletionMessage(content="dummy", role="assistant"),
        )
    ]
    completion = ChatCompletion(
        id="dummy_id",
        choices=choices,
        created=0,
        model="dummy_model",
        object="chat.completion",
    )
    messages = [
        {"content": [{"type": "image_url", "url": "http://example.com/img1.jpg"}]},
        {"content": [{"type": "image_url"}]},
        {"content": "not a list"},  # Should be skipped because content is not a list
        {
            "no_content": [{"type": "image_url", "url": "http://example.com/img2.jpg"}]
        },  # Skipped because no "content" key
        {
            "content": [
                {"type": "other"},
                {"type": "image_url", "url": "http://example.com/img3.jpg"},
            ]
        },
    ]
    call_response = GroqCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=messages,
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    cost_md = call_response.cost_metadata
    assert hasattr(cost_md, "images")
    assert cost_md.images == [
        ImageMetadata(width=0, height=0),
        ImageMetadata(width=0, height=0),
        ImageMetadata(width=0, height=0),
    ]



"""Tests the `gemini._utils.get_json_output` module."""

import pytest
from google.ai.generativelanguage import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
    Part,
)
from google.generativeai.types import (  # type: ignore
    GenerateContentResponse as GenerateContentResponseType,
)

from mirascope.core.gemini._utils._get_json_output import get_json_output
from mirascope.core.gemini.call_response import GeminiCallResponse
from mirascope.core.gemini.call_response_chunk import GeminiCallResponseChunk


@pytest.fixture()
def mock_generate_content_response() -> GenerateContentResponseType:
    """Returns a mock `GenerateContentResponse` instance."""
    return GenerateContentResponseType.from_response(
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=1,
                    content=Content(
                        parts=[
                            Part(text='{"key": "value"}'),
                            Part(
                                function_call=FunctionCall(
                                    name="FormatBook",
                                    args={
                                        "title": "The Name of the Wind",
                                        "author": "Patrick Rothfuss",
                                    },
                                )
                            ),
                        ],
                        role="model",
                    ),
                )
            ]
        )
    )


def test_get_json_output_call_response(
    mock_generate_content_response: GenerateContentResponseType,
) -> None:
    """Tests the `get_json_output` function with a call response."""
    call_response = GeminiCallResponse(
        metadata={},
        response=mock_generate_content_response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == '{"key": "value"}'
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    mock_generate_content_response.candidates[0].content.parts.pop(1)
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk(
    mock_generate_content_response: GenerateContentResponseType,
) -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    call_response_chunk = GeminiCallResponseChunk(chunk=mock_generate_content_response)
    assert get_json_output(call_response_chunk, json_mode=True) == '{"key": "value"}'

    with pytest.raises(
        ValueError, match="Gemini only supports structured streaming in json mode."
    ):
        get_json_output(call_response_chunk, json_mode=False)



"""The Mirascope Google Module."""

from typing import TypeAlias

from google.genai.types import ContentDict, FunctionResponse

from ..base import BaseMessageParam
from ._call import google_call
from ._call import google_call as call
from .call_params import GoogleCallParams
from .call_response import GoogleCallResponse
from .call_response_chunk import GoogleCallResponseChunk
from .dynamic_config import GoogleDynamicConfig
from .stream import GoogleStream
from .tool import GoogleTool

GoogleMessageParam: TypeAlias = ContentDict | FunctionResponse | BaseMessageParam

__all__ = [
    "GoogleCallParams",
    "GoogleCallResponse",
    "GoogleCallResponseChunk",
    "GoogleDynamicConfig",
    "GoogleMessageParam",
    "GoogleStream",
    "GoogleTool",
    "call",
    "google_call",
]



"""This module defines the function return type for functions as LLM calls."""

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)
from anthropic.types import MessageParam

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import AnthropicCallParams

AnthropicDynamicConfig = BaseDynamicConfig[
    MessageParam | BaseMessageParam,
    AnthropicCallParams,
    Anthropic | AnthropicBedrock | AnthropicVertex,
]
AsyncAnthropicDynamicConfig = BaseDynamicConfig[
    MessageParam | BaseMessageParam,
    AnthropicCallParams,
    AsyncAnthropic | AsyncAnthropicBedrock | AsyncAnthropicVertex,
]
"""The function return type for functions wrapped with the `anthropic_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.anthropic import AnthropicDynamicConfig, anthropic_call


@anthropic_call("claude-3-5-sonnet-20240620")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> AnthropicDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""MCP client utility functions."""

import base64
import string
from collections.abc import AsyncGenerator, Awaitable, Callable
from contextlib import asynccontextmanager
from typing import Any

from anyio import create_memory_object_stream, create_task_group
from anyio.streams.memory import MemoryObjectReceiveStream
from mcp.types import (
    EmbeddedResource,
    ImageContent,
    JSONRPCMessage,
    PromptMessage,
    TextContent,
    TextResourceContents,
)
from mcp.types import Tool as MCPTool
from pydantic import BaseModel, ConfigDict, Field, create_model

from ..core import BaseMessageParam, BaseTool
from ..core.base import AudioPart, DocumentPart, ImagePart

_UNSET = object()


def create_tool_call(
    name: str,
    call_tool: Callable[[str, dict | None], Awaitable[Any]],
) -> Callable[..., Awaitable[list[str | ImageContent | EmbeddedResource]]]:
    """Create a tool call function for a Mirascope Tool.

    Args:
        name: The name of the tool
        call_tool: The function to call the tool

    Returns:
        A function that can be used as the call method for a Tool
    """

    async def call(self: BaseTool) -> list[str | ImageContent | EmbeddedResource]:
        args = {key: value for key, value in self.args.items() if value is not _UNSET}
        result = await call_tool(name, args if args else None)
        if result.isError:
            raise RuntimeError(f"MCP Server returned error: {self._name()}")
        return [
            content.text if isinstance(content, TextContent) else content
            for content in result.content
        ]

    return call


def convert_prompt_message_to_base_message_params(
    prompt_message: PromptMessage,
) -> BaseMessageParam:
    """
    Convert a single PromptMessage back into a BaseMessageParam.

    Args:
        prompt_message: A PromptMessage instance.

    Returns:
        A BaseMessageParam instance representing the given prompt_message.

    Raises:
        ValueError: If the role is invalid or if the content type is unsupported.
    """

    # Validate role
    if prompt_message.role not in ["user", "assistant"]:
        raise ValueError(f"invalid role: {prompt_message.role}")

    content = prompt_message.content

    # Handle TextContent
    if isinstance(content, TextContent):
        # If it's text, we can just return a single string
        return BaseMessageParam(role=prompt_message.role, content=content.text)

    # Handle ImageContent
    elif isinstance(content, ImageContent):
        decoded_image = base64.b64decode(content.data)

        image_part = ImagePart(
            type="image",
            image=decoded_image,
            media_type=content.mimeType,
            detail=None,  # detail not provided by PromptMessage
        )
        return BaseMessageParam(role=prompt_message.role, content=[image_part])

    elif isinstance(content, EmbeddedResource):
        resource = content.resource
        if isinstance(resource, TextResourceContents):
            # For text content embedded resources, just return the text
            return BaseMessageParam(role=prompt_message.role, content=resource.text)
        else:
            mime_type = resource.mimeType
            if not mime_type:
                raise ValueError(
                    "BlobResourceContents has no mimeType, cannot determine content type."
                )

            decoded_data = base64.b64decode(resource.blob)

            if mime_type.startswith("image/"):
                # Treat as ImagePart
                image_part = ImagePart(
                    type="image",
                    image=decoded_data,
                    media_type=mime_type,
                    detail=None,
                )
                return BaseMessageParam(role=prompt_message.role, content=[image_part])
            elif mime_type == "application/pdf":
                doc_part = DocumentPart(
                    type="document", media_type=mime_type, document=decoded_data
                )
                return BaseMessageParam(role=prompt_message.role, content=[doc_part])
            elif mime_type.startswith("audio/"):
                # Treat as AudioPart
                audio_part = AudioPart(
                    type="audio", media_type=mime_type, audio=decoded_data
                )
                return BaseMessageParam(role=prompt_message.role, content=[audio_part])
            else:
                raise ValueError(f"Unsupported mime type: {mime_type}")

    else:
        raise ValueError(f"Unsupported content type: {type(content)}")


def snake_to_pascal(snake: str) -> str:
    """Convert a snake_case string to PascalCase.

    Args:
        snake: A snake_case string

    Returns:
        The string converted to PascalCase
    """
    return string.capwords(snake.replace("_", " ")).replace(" ", "")


def build_object_type(
    properties: dict[str, Any], required: list[str], name: str
) -> type[BaseModel]:
    """Build a Pydantic model from JSON Schema properties.

    Args:
        properties: JSON Schema properties dictionary
        required: List of required property names
        name: Name for the generated model

    Returns:
        A dynamically created Pydantic model class
    """
    fields = {}
    for prop_name, prop_schema in properties.items():
        class_name = snake_to_pascal(prop_name)
        type_ = json_schema_to_python_type(prop_schema, class_name)
        if prop_name in required:
            fields[prop_name] = (
                type_,
                Field(..., description=prop_schema.get("description")),
            )
        else:
            fields[prop_name] = (
                type_ | None,
                Field(None, description=prop_schema.get("description")),
            )

    return create_model(name, __config__=ConfigDict(extra="allow"), **fields)


def json_schema_to_python_type(schema: dict[str, Any], name: str = "Model") -> Any:  # noqa: ANN401
    """
    Recursively convert a JSON Schema snippet into a Python type annotation or a dynamically generated pydantic model.

    Args:
        schema: JSON Schema to convert
        name: Name for any generated models

    Returns:
        Corresponding Python type or Pydantic model
    """
    json_type = schema.get("type", "any")

    if json_type == "string":
        return str
    elif json_type == "number":
        return float
    elif json_type == "integer":
        return int
    elif json_type == "boolean":
        return bool
    elif json_type == "array":
        # Recursively determine the items type for arrays
        items_schema = schema.get("items", {})
        items_type = json_schema_to_python_type(items_schema)
        return list[items_type]
    elif json_type == "object":
        # Recursively build a dynamic model for objects
        properties = schema.get("properties", {})
        required = schema.get("required", [])
        return build_object_type(properties, required, name)
    else:
        # Default fallback
        return Any


def create_tool_from_mcp_tool(tool: MCPTool) -> type[BaseTool]:
    """
    Create a `BaseTool` type definition from the given MCP Tool instance.

    Args:
        tool: MCP tool instance.

    Returns:
        A dynamically created `Tool` schema.
    """
    schema = tool.inputSchema
    properties = schema.get("properties", {})
    required_fields = schema.get("required", [])

    fields = {}
    for field_name, field_schema in properties.items():
        field_type = json_schema_to_python_type(field_schema)
        if field_schema.get("nullable", False):
            field_type = field_type | None

        field_info = {
            "default": field_schema.get(
                "default", ... if field_name in required_fields else _UNSET
            )
        }
        if description := field_schema.get("description", ""):
            field_info["description"] = description

        fields[field_name] = (field_type, Field(**field_info))

    return create_model(snake_to_pascal(tool.name), __base__=BaseTool, **fields)


@asynccontextmanager
async def read_stream_exception_filer(
    original_read: MemoryObjectReceiveStream[JSONRPCMessage | Exception],
    exception_handler: Callable[[Exception], None] | None = None,
) -> AsyncGenerator[MemoryObjectReceiveStream[JSONRPCMessage], None]:
    """
    Handle exceptions in the original stream.
    Default handler is to ignore read stream exception for invalid JSON lines and log them.

    Args:
        original_read: The original stream to read from.
        exception_handler: An optional handler for exception.

    Yields:
        A filtered stream with exceptions removed
    """
    read_stream_writer, filtered_read = create_memory_object_stream(0)

    async def task() -> None:
        try:
            async with original_read:
                async for msg in original_read:
                    if isinstance(msg, Exception):
                        if exception_handler:
                            exception_handler(msg)
                        continue
                    await read_stream_writer.send(msg)
        finally:
            await read_stream_writer.aclose()

    async with create_task_group() as tg:
        tg.start_soon(task)
        try:
            yield filtered_read
        finally:
            await filtered_read.aclose()
            tg.cancel_scope.cancel()



"""Tests the `_utils.parse_content_template` function."""

from io import BytesIO
from unittest.mock import MagicMock, patch

import pytest
from PIL import Image

from mirascope.core.base._utils._parse_content_template import parse_content_template
from mirascope.core.base.message_param import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    DocumentPart,
    ImagePart,
    ImageURLPart,
    TextPart,
)


def test_parse_content_template() -> None:
    """Test the parse_content_template function."""
    assert parse_content_template("user", "", {}) is None
    assert parse_content_template("user", " ", {}) is None
    assert parse_content_template("system", "", {}) is None
    assert parse_content_template("system", " \u200b ", {}) is None
    template = "This is a {var1} template with {var2} variables."
    values = {"var1": "test", "var2": "two"}
    expected = BaseMessageParam(
        role="user", content="This is a test template with two variables."
    )
    assert parse_content_template("user", template, values) == expected


@pytest.fixture
def mock_jpeg_bytes() -> bytes:
    return b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\x00\x00\xff\xdb\x00C\x00\x08\x06\x06\x07\x06\x05\x08\x07\x07\x07\t\t\x08\n\x0c\x14\r\x0c\x0b\x0b\x0c\x19\x12\x13\x0f\x14\x1d\x1a\x1f\x1e\x1d\x1a\x1c\x1c $.' \",#\x1c\x1c(7),01444\x1f'9=82<.342\xff\xc0\x00\x0b\x08\x00\x01\x00\x01\x01\x01\x11\x00\xff\xc4\x00\x1f\x00\x00\x01\x05\x01\x01\x01\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\xff\xc4\x00\xb5\x10\x00\x02\x01\x03\x03\x02\x04\x03\x05\x05\x04\x04\x00\x00\x01}\x01\x02\x03\x00\x04\x11\x05\x12!1A\x06\x13Qa\x07\"q\x142\x81\x91\xa1\x08#B\xb1\xc1\x15R\xd1\xf0$3br\x82\t\n\x16\x17\x18\x19\x1a%&'()*456789:CDEFGHIJSTUVWXYZcdefghijstuvwxyz\x83\x84\x85\x86\x87\x88\x89\x8a\x92\x93\x94\x95\x96\x97\x98\x99\x9a\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xff\xda\x00\x08\x01\x01\x00\x00?\x00\xf9\xfe\xbf\xff\xd9"


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_images(
    mock_urlopen: MagicMock, mock_open: MagicMock, mock_jpeg_bytes: bytes
) -> None:
    """Test the parse_content_template function with image templates."""
    mock_response = MagicMock()
    mock_response.read = lambda: mock_jpeg_bytes
    mock_urlopen.return_value.__enter__.return_value = mock_response
    mock_open.return_value.__enter__.return_value = mock_response
    template = "Analyze this image: {url:image}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this image:"),
            ImageURLPart(type="image_url", url="https://", detail=None),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this image:"),
            ImagePart(
                type="image",
                media_type="image/jpeg",
                image=mock_jpeg_bytes,
                detail=None,
            ),
        ],
    )
    assert parse_content_template("user", template, {"url": "./image.jpg"}) == expected
    assert (
        parse_content_template("user", template, {"url": mock_jpeg_bytes}) == expected
    )
    assert (
        parse_content_template(
            "user", template, {"url": Image.open(BytesIO(mock_jpeg_bytes))}
        )
        == expected
    )

    template = "Analyze this image: {url:image(detail=low)}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this image:"),
            ImageURLPart(type="image_url", url="https://", detail="low"),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected

    template = "Analyze these images: {urls:images}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these images:"),
            ImageURLPart(type="image_url", url="https://", detail=None),
            ImageURLPart(type="image_url", url="https://.", detail=None),
        ],
    )
    assert (
        parse_content_template("user", template, {"urls": ["https://", "https://."]})
        == expected
    )
    expected = expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these images:"),
            ImagePart(
                type="image",
                media_type="image/jpeg",
                image=mock_jpeg_bytes,
                detail=None,
            ),
            ImagePart(
                type="image",
                media_type="image/jpeg",
                image=mock_jpeg_bytes,
                detail=None,
            ),
        ],
    )
    assert (
        parse_content_template(
            "user", template, {"urls": ["./image.jpg", "./image.jpg"]}
        )
        == expected
    )
    assert (
        parse_content_template(
            "user", template, {"urls": [mock_jpeg_bytes, mock_jpeg_bytes]}
        )
        == expected
    )

    with pytest.raises(
        ValueError,
        match="When using 'images' template, 'urls' must be a list.",
    ):
        parse_content_template("user", template, {"urls": None})


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_audio(
    mock_urlopen: MagicMock, mock_open: MagicMock
) -> None:
    """Test the parse_content_template function with image templates."""
    audio_data = b"ID3audio data"
    mock_response = MagicMock()
    mock_response.read = lambda: audio_data
    mock_urlopen.return_value.__enter__.return_value = mock_response
    mock_open.return_value.__enter__.return_value = mock_response
    template = "Analyze this audio: {url:audio}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this audio:"),
            AudioURLPart(type="audio_url", url="https://"),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this audio:"),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
        ],
    )
    assert parse_content_template("user", template, {"url": "./audio.mp3"}) == expected
    assert parse_content_template("user", template, {"url": audio_data}) == expected

    template = "Analyze these audio files: {urls:audios}"

    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these audio files:"),
            AudioURLPart(type="audio_url", url="https://"),
            AudioURLPart(type="audio_url", url="https://."),
        ],
    )
    assert (
        parse_content_template("user", template, {"urls": ["https://", "https://."]})
        == expected
    )
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these audio files:"),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
            AudioPart(type="audio", media_type="audio/mp3", audio=audio_data),
        ],
    )
    assert (
        parse_content_template(
            "user", template, {"urls": ["./audio.mp3", "./audio.mp3"]}
        )
        == expected
    )
    assert (
        parse_content_template("user", template, {"urls": [audio_data, audio_data]})
        == expected
    )

    with pytest.raises(
        ValueError,
        match="When using 'audios' template, 'urls' must be a list.",
    ):
        parse_content_template("user", template, {"urls": None})


def test_parse_content_template_cache_control() -> None:
    """Test the parse_content_template function with a cache_control part."""
    expected = BaseMessageParam(
        role="user",
        content=[CacheControlPart(type="cache_control", cache_type="ephemeral")],
    )
    assert parse_content_template("user", "{:cache_control}", {}) == expected
    assert (
        parse_content_template("user", "{:cache_control(type=ephemeral)}", {})
        == expected
    )
    assert parse_content_template("user", "{not_used:cache_control}", {}) == expected

    expected.role = "system"
    assert parse_content_template("system", "{:cache_control}", {}) == expected


def test_parse_content_template_texts() -> None:
    """Test the parse_content_template function with text templates."""
    # Test single text template
    template = "Process this text: {content:text}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process this text:"),
            TextPart(type="text", text="Hello world"),
        ],
    )
    assert (
        parse_content_template("user", template, {"content": "Hello world"}) == expected
    )

    # Test texts template with list
    template = "Process these texts: {content:texts}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process these texts:"),
            TextPart(type="text", text="Hello"),
            TextPart(type="text", text="world"),
        ],
    )
    assert (
        parse_content_template("user", template, {"content": ["Hello", "world"]})
        == expected
    )

    # Test empty list
    assert parse_content_template(
        "user", template, {"content": []}
    ) == BaseMessageParam(role="user", content="Process these texts:")

    # Test texts validation
    with pytest.raises(
        ValueError,
        match="When using 'texts' template, 'content' must be a list.",
    ):
        parse_content_template("user", template, {"content": "not a list"})

    # Test attribute access
    template = "Process this text: {attr:text}"
    assert parse_content_template(
        "user", template, {"attr": "value"}
    ) == BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process this text:"),
            TextPart(type="text", text="value"),
        ],
    )

    # Test empty text part
    assert parse_content_template("user", template, {"attr": ""}) == BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process this text:"),
            TextPart(type="text", text=""),
        ],
    )


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_document(
    mock_urlopen: MagicMock, mock_open: MagicMock
) -> None:
    """Test the parse_content_template function with document templates."""
    document_data = b"%PDFdocument data"  # Magic bytes for PDF files
    mock_response = MagicMock()
    mock_response.read = lambda: document_data
    mock_urlopen.return_value.__enter__.return_value = mock_response
    mock_open.return_value.__enter__.return_value = mock_response

    # Test single document input
    template = "Analyze this document: {url:document}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze this document:"),
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=document_data,
            ),
        ],
    )
    assert parse_content_template("user", template, {"url": "https://"}) == expected
    assert parse_content_template("user", template, {"url": "./doc.pdf"}) == expected
    assert parse_content_template("user", template, {"url": document_data}) == expected

    # Test multiple document inputs
    template = "Analyze these documents: {urls:documents}"
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Analyze these documents:"),
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=document_data,
            ),
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=document_data,
            ),
        ],
    )
    assert (
        parse_content_template("user", template, {"urls": ["https://", "https://."]})
        == expected
    )
    assert (
        parse_content_template("user", template, {"urls": ["./doc.pdf", "./doc.pdf"]})
        == expected
    )
    assert (
        parse_content_template(
            "user", template, {"urls": [document_data, document_data]}
        )
        == expected
    )

    # Test error case for invalid input type
    with pytest.raises(
        ValueError,
        match="When using 'documents' template, 'urls' must be a list.",
    ):
        parse_content_template("user", template, {"urls": None})


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_parts(
    mock_urlopen: MagicMock, mock_open: MagicMock, mock_jpeg_bytes: bytes
) -> None:
    """Test the parse_content_template function with parts templates."""
    # Test single parts template
    template = "Process these parts: {content:parts}"
    mixed_parts = [
        TextPart(type="text", text="Hello world"),
        ImagePart(
            type="image",
            media_type="image/jpeg",
            image=mock_jpeg_bytes,
            detail=None,
        ),
        AudioPart(
            type="audio",
            media_type="audio/mp3",
            audio=b"audio_bytes",
        ),
    ]
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process these parts:"),
            *mixed_parts,
        ],
    )
    assert (
        parse_content_template("user", template, {"content": mixed_parts}) == expected
    )

    # Test empty list
    assert parse_content_template(
        "user", template, {"content": []}
    ) == BaseMessageParam(role="user", content="Process these parts:")

    # Test parts validation
    with pytest.raises(
        ValueError,
        match="When using 'parts' template, 'content' must be a list.",
    ):
        parse_content_template("user", template, {"content": "not a list"})

    with pytest.raises(
        ValueError,
        match="When using 'parts' template, 'content' must be a list of valid content parts.",
    ):
        parse_content_template("user", template, {"content": ["invalid part"]})


@patch(
    "mirascope.core.base._utils._parse_content_template.open", new_callable=MagicMock
)
@patch("urllib.request.urlopen", new_callable=MagicMock)
def test_parse_content_template_part(
    mock_urlopen: MagicMock, mock_open: MagicMock
) -> None:
    """Test the parse_content_template function with a single part template."""
    # Test single part template
    template = "Process this part: {content:part}"
    text_part = TextPart(type="text", text="Hello world")
    expected = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Process this part:"),
            text_part,
        ],
    )
    assert parse_content_template("user", template, {"content": text_part}) == expected

    # Test validation
    with pytest.raises(
        ValueError,
        match="When using 'part' template, 'content' must be a valid content part.",
    ):
        parse_content_template("user", template, {"content": "not a part"})



from unittest.mock import PropertyMock, patch

import pytest
from google.cloud.aiplatform_v1beta1.types import content as gapic_content_types
from google.cloud.aiplatform_v1beta1.types import tool as gapic_tool_types
from vertexai.generative_models import Content, Part

from mirascope.core.base import ImagePart
from mirascope.core.base.message_param import (
    BaseMessageParam,
    DocumentPart,
    ImageURLPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.vertex._utils._message_param_converter import (
    VertexMessageParamConverter,
    _to_document_part,
    _to_image_part,
)


def test_vertex_convert_parts_image():
    raw_gapic_part = gapic_content_types.Part(
        inline_data=gapic_content_types.Blob(
            mime_type="image/png", data=b"\x89PNG\r\n\x1a\n"
        )
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    with patch.object(Part, "text", new_callable=PropertyMock(return_value="")):
        results = VertexMessageParamConverter.from_provider([message])
        assert len(results) == 1
        result = results[0]
        assert isinstance(result.content, list)
        assert len(result.content) == 1
        assert isinstance(result.content[0], ImagePart)
        assert result.content[0].media_type == "image/png"
        assert result.content[0].image == b"\x89PNG\r\n\x1a\n"
        assert result.role == "assistant"


def test_vertex_convert_parts_image_url():
    raw_gapic_part = gapic_content_types.Part(
        file_data=gapic_content_types.FileData(
            mime_type="image/png", file_uri="https://example.com/image.png"
        )
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    with patch.object(Part, "text", new_callable=PropertyMock(return_value="")):
        results = VertexMessageParamConverter.from_provider([message])
        assert len(results) == 1
        result = results[0]
        assert isinstance(result.content, list)
        assert len(result.content) == 1
        assert isinstance(result.content[0], ImageURLPart)
        assert result.content[0].url == "https://example.com/image.png"
        assert result.role == "assistant"


def test_vertex_convert_parts_unsupported_image():
    raw_gapic_part = gapic_content_types.Part(
        inline_data=gapic_content_types.Blob(mime_type="image/tiff", data=b"fake")
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    with (
        patch.object(Part, "text", new_callable=PropertyMock(return_value="")),
        pytest.raises(
            ValueError,
            match="Unsupported inline_data mime type: image/tiff. Cannot convert to BaseMessageParam.",
        ),
    ):
        VertexMessageParamConverter.from_provider([message])


def test_vertex_convert_parts_inline_data_pdf():
    raw_gapic_part = gapic_content_types.Part(
        inline_data=gapic_content_types.Blob(
            mime_type="application/pdf", data=b"%PDF-1.4..."
        )
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    results = VertexMessageParamConverter.from_provider([message])
    assert len(results) == 1
    assert results == [
        BaseMessageParam(
            role="assistant",
            content=[
                DocumentPart(
                    type="document",
                    media_type="application/pdf",
                    document=b"%PDF-1.4...",
                )
            ],
        )
    ]


def test_vertex_convert_parts_unsupported_inline_data():
    raw_gapic_part = gapic_content_types.Part(
        inline_data=gapic_content_types.Blob(mime_type="application/json", data=b"{}")
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    with (
        patch.object(Part, "text", new_callable=PropertyMock(return_value="")),
        pytest.raises(
            ValueError,
            match="Unsupported inline_data mime type: application/json. Cannot convert to BaseMessageParam.",
        ),
    ):
        VertexMessageParamConverter.from_provider([message])


def test_to_image_part_unsupported_image():
    with pytest.raises(
        ValueError,
        match="Unsupported image media type: application/json. Expected one of: image/jpeg, image/png, image/gif, image/webp.",
    ):
        _to_image_part(mime_type="application/json", data=b"{}")


def test_to_document_part_unsupported_document():
    with pytest.raises(
        ValueError,
        match="Unsupported document media type: application/json. Only application/pdf is supported.",
    ):
        _to_document_part(mime_type="application/json", data=b"{}")


def test_vertex_convert_parts_tool_result():
    raw_gapic_part = gapic_content_types.Part(
        function_call=gapic_tool_types.FunctionCall(
            name="test", args={"arg1": "value1"}
        )
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    with patch.object(Part, "text", new_callable=PropertyMock(return_value="")):
        results = VertexMessageParamConverter.from_provider([message])
        assert len(results) == 1
        result = results[0]
        assert result.role in ("assistant", "tool")
        assert len(result.content) == 1
        part_call = result.content[0]
        assert isinstance(part_call, ToolCallPart)
        assert part_call.type == "tool_call"
        assert part_call.name == "test"
        assert part_call.args == {"arg1": "value1"}


def test_gemini_convert_parts_document():
    """
    If file_data is a PDF, produce a DocumentPart.
    """
    part = gapic_content_types.Part(
        file_data=gapic_content_types.FileData(
            mime_type="application/pdf", file_uri=b"%PDF-1.4..."
        )
    )
    with pytest.raises(
        ValueError,
        match='FileData.file_uri is not support: mime_type: "application/pdf"\nfile_uri: "%PDF-1.4..."\n. Cannot convert to BaseMessageParam.',
    ):
        VertexMessageParamConverter.from_provider(
            [Content(role="user", parts=[Part._from_gapic(part)])]
        )


def test_to_provider():
    """
    If there's no content but tool_calls, we only get ToolCallParts in the final list.
    """
    message_param = Content(
        role="assistant",
        parts=[
            Part.from_text("This is a message"),
        ],
    )
    results = VertexMessageParamConverter.to_provider([message_param])  # pyright: ignore [reportArgumentType]
    assert len(results) == 1
    assert results[0].role == "assistant"
    assert results[0].parts[0].text == "This is a message"


def test_function_response():
    raw_gapic_part = gapic_content_types.Part(
        function_response=gapic_tool_types.FunctionResponse(
            name="test", response={"result": "value"}
        )
    )
    part = Part._from_gapic(raw_gapic_part)
    message = Content(role="assistant", parts=[part])
    results = VertexMessageParamConverter.from_provider([message])
    assert len(results) == 1
    assert results == [
        BaseMessageParam(
            role="assistant",
            content=[
                ToolResultPart(
                    type="tool_result",
                    name="test",
                    content="value",
                    id=None,
                    is_error=False,
                )
            ],
        )
    ]



"""This module contains the setup_call function for OpenAI tools."""

import inspect
import warnings
from collections.abc import Awaitable, Callable
from typing import Any, cast, overload

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import (
    DEFAULT_TOOL_DOCSTRING,
    AsyncCreateFn,
    CreateFn,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import OpenAICallKwargs
from ..call_params import OpenAICallParams
from ..dynamic_config import AsyncOpenAIDynamicConfig, OpenAIDynamicConfig
from ..tool import GenerateOpenAIStrictToolJsonSchema, OpenAITool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: AsyncOpenAI | AsyncAzureOpenAI | None,
    fn: Callable[..., Awaitable[AsyncOpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncOpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: OpenAI | AzureOpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: OpenAI | AsyncOpenAI | AzureOpenAI | AsyncAzureOpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig]
    | Callable[..., Awaitable[AsyncOpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig | AsyncOpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk]
    | AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        OpenAITool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(OpenAICallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | ChatCompletionMessageParam], messages)
    messages = convert_message_params(messages)
    if json_mode:
        if response_model and response_model.model_config.get("strict", False):
            call_kwargs["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "name": response_model.__name__,
                    "description": response_model.__doc__ or DEFAULT_TOOL_DOCSTRING,
                    "strict": True,
                    "schema": response_model.model_json_schema(
                        schema_generator=GenerateOpenAIStrictToolJsonSchema
                    ),
                },
            }
        else:
            call_kwargs["response_format"] = {"type": "json_object"}
            json_mode_content = _utils.json_mode_content(response_model).strip()
            messages.append(
                ChatCompletionUserMessageParam(role="user", content=json_mode_content)
            )
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        if tool_types and tool_types[0].model_config.get("strict", False):
            warnings.warn(
                "You must set `json_mode=True` to use `strict=True` response models. "
                "Ignoring `strict` and using tools for extraction.",
                UserWarning,
            )
        call_kwargs["tool_choice"] = "required"
    if stream:
        call_kwargs["stream_options"] = (
            {"include_usage": True}
            if "stream_options" not in call_kwargs
            else call_kwargs["stream_options"].update({"include_usage": True})  # pyright: ignore [reportOptionalMemberAccess]
        )
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = AsyncOpenAI() if inspect.iscoroutinefunction(fn) else OpenAI()
    create = (
        get_async_create_fn(client.chat.completions.create)
        if isinstance(client, AsyncOpenAI)
        else get_create_fn(client.chat.completions.create)
    )
    return create, prompt_template, messages, tool_types, call_kwargs



from typing import Any, cast
from unittest.mock import MagicMock, patch

import pytest
from PIL import Image

from mirascope.core import BaseMessageParam
from mirascope.core.base import ImagePart, Messages, TextPart
from mirascope.core.base._utils._messages_decorator import messages_decorator
from mirascope.core.base.dynamic_config import DynamicConfigMessages


def test_messages_decorator_str_return():
    @messages_decorator()
    def recommend_book(genre: str) -> str:
        return f"recommend a {genre} book"

    result = recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "user"
    assert result[0].content == "recommend a fantasy book"


def test_messages_decorator_dynamic_configuration_return():
    @messages_decorator()
    def recommend_book(genre: str) -> DynamicConfigMessages:
        return cast(DynamicConfigMessages, {"messages": f"recommend a {genre} book"})

    result = recommend_book("fantasy")
    assert isinstance(result, dict)
    assert "messages" in result
    assert result["messages"] == "recommend a fantasy book"


def test_messages_decorator_with_none_argument_str_return():
    @messages_decorator()
    def recommend_book(genre: str) -> str:
        return f"recommend a {genre} book"

    result = recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "user"
    assert result[0].content == "recommend a fantasy book"


@pytest.mark.asyncio
async def test_messages_decorator_str_return_async():
    @messages_decorator()
    async def recommend_book(genre: str) -> str:
        return f"recommend a {genre} book"

    result = await recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "user"
    assert result[0].content == "recommend a fantasy book"


@pytest.mark.asyncio
async def test_prompt_messages_decorator_configuration_return_async():
    @messages_decorator()
    async def recommend_book(genre: str) -> DynamicConfigMessages:
        return cast(DynamicConfigMessages, {"messages": f"recommend a {genre} book"})

    result = cast(DynamicConfigMessages, await recommend_book("fantasy"))
    assert isinstance(result, dict)
    assert "messages" in result
    assert result["messages"] == "recommend a fantasy book"


def test_list_str_return():
    @messages_decorator()
    def recommend_book(genre: str) -> list[str]:
        return ["hello!", f"recommend a {genre} book"]

    result = recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result == [
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="hello!"),
                TextPart(type="text", text="recommend a fantasy book"),
            ],
        )
    ]


def test_list_message_role_return():
    @messages_decorator()
    def recommend_book(genre: str) -> Messages.Type:
        return [
            Messages.System("You are a librarian"),
            Messages.Assistant("I can help you find a book"),
            Messages.User(
                content=[
                    TextPart(text="recommend a", type="text"),
                    TextPart(text=genre, type="text"),
                    TextPart(text="book", type="text"),
                ]
            ),
        ]

    result = recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 3
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "system"
    assert result[0].content == "You are a librarian"
    assert result[1].role == "assistant"
    assert result[1].content == "I can help you find a book"
    assert isinstance(result[2], BaseMessageParam)
    assert result[2].role == "user"
    assert result[2].content == [
        TextPart(type="text", text="recommend a"),
        TextPart(type="text", text="fantasy"),
        TextPart(type="text", text="book"),
    ]


def test_base_message_param_return():
    @messages_decorator()
    def recommend_book(genre: str) -> BaseMessageParam:
        return BaseMessageParam(role="user", content=f"hello! recommend a {genre} book")

    result = recommend_book("fantasy")
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "user"
    assert result[0].content == "hello! recommend a fantasy book"


@patch(
    "mirascope.core.base._utils._convert_messages_to_message_params.pil_image_to_bytes",
    new_callable=MagicMock,
)
def test_multimodal_return(mock_pil_image_to_bytes: MagicMock):
    mock_pil_image_to_bytes.return_value = b"image_bytes"
    mock_image = MagicMock(spec=Image.Image)
    mock_image.format = "PNG"

    @messages_decorator()
    def recommend_book(previous_book: Image.Image) -> list[Any]:
        return ["I just read this book:", previous_book, "What should I read next?"]

    result = recommend_book(mock_image)

    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], BaseMessageParam)
    assert result[0].role == "user"
    assert len(result[0].content) == 3
    assert result[0].content[0] == TextPart(type="text", text="I just read this book:")
    assert result[0].content[1] == ImagePart(
        type="image", media_type="image/png", image=b"image_bytes", detail=None
    )
    assert result[0].content[2] == TextPart(
        type="text", text="What should I read next?"
    )



"""Utility for converting `BaseMessageParam` to `ChatRequestMessage`."""

import base64
import json
from typing import cast

from azure.ai.inference.models import (
    AssistantMessage,
    ChatCompletionsToolCall,
    ChatRequestMessage,
    FunctionCall,
    ToolMessage,
    UserMessage,
)

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ChatRequestMessage],
) -> list[ChatRequestMessage]:
    converted_message_params: list[ChatRequestMessage | ToolMessage] = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(UserMessage(content=content))
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.model_dump())
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. Azure"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    data = base64.b64encode(part.image).decode("utf-8")
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:{part.media_type};base64,{data}",
                                "detail": part.detail if part.detail else "auto",
                            },
                        }
                    )
                elif part.type == "image_url":
                    converted_content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": part.url,
                                "detail": part.detail if part.detail else "auto",
                            },
                        }
                    )
                elif part.type == "tool_call":
                    converted_message_param = AssistantMessage(
                        tool_calls=[
                            ChatCompletionsToolCall(
                                id=part.id,  # pyright: ignore [reportArgumentType]
                                function=FunctionCall(
                                    name=part.name, arguments=json.dumps(part.args)
                                ),
                            )
                        ]
                    )

                    if converted_content:
                        if len(converted_content) == 1:
                            if converted_content[0]["type"] == "text":
                                converted_message_param["content"] = converted_content[
                                    0
                                ]["text"]
                        else:
                            converted_message_params.append(
                                ChatRequestMessage(
                                    {
                                        "role": message_param.role,
                                        "content": converted_content,
                                    }
                                )
                            )
                        converted_content = []
                    converted_message_params.append(converted_message_param)
                elif part.type == "tool_result":
                    if converted_content:
                        converted_message_params.append(
                            ChatRequestMessage(
                                {
                                    "role": message_param.role,
                                    "content": converted_content,
                                }
                            )
                        )
                        converted_content = []
                    converted_message_params.append(
                        ToolMessage(
                            content=part.content,
                            tool_call_id=cast(str, part.id),
                        )
                    )
                else:
                    raise ValueError(
                        "Azure currently only supports text and image parts. "
                        f"Part provided: {part.type}"
                    )
            if converted_content:
                converted_message_params.append(
                    ChatRequestMessage(
                        {"role": message_param.role, "content": converted_content}
                    )
                )
    return converted_message_params



import inspect
from functools import cached_property
from typing import Any

from pydantic._internal._model_construction import ModelMetaclass


class _ResponseMetaclass(ModelMetaclass):
    def __new__(
        mcls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,  # noqa: ANN401
    ) -> type:
        cls = super().__new__(mcls, name, bases, namespace)
        cls.__abstractmethods__ = frozenset()
        cls._properties = []

        for n, v in inspect.getmembers(cls):
            if isinstance(v, property):
                f = v.fget
            elif isinstance(v, cached_property):
                f = v.func
            else:
                continue

            if not getattr(f, "__isabstractmethod__", False):
                cls._properties.append(n)

        return cls



"""Tests the `openai._utils.setup_call` module."""

from typing import ClassVar
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base import ResponseModelConfigDict
from mirascope.core.openai._utils._convert_common_call_params import (
    convert_common_call_params,
)
from mirascope.core.openai._utils._setup_call import setup_call
from mirascope.core.openai.tool import OpenAITool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch("mirascope.core.openai._utils._setup_call.OpenAI", new_callable=MagicMock)
@patch(
    "mirascope.core.openai._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.openai._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_openai: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_utils.setup_call = mock_base_setup_call
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "model" in call_kwargs and call_kwargs["model"] == "gpt-4o"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    assert "stream_options" not in call_kwargs
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, OpenAITool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    mock_create = mock_openai.return_value.chat.completions.create
    assert create(**call_kwargs)
    mock_create.assert_called_once_with(**call_kwargs)
    mock_create.reset_mock()


@patch("mirascope.core.openai._utils._setup_call.OpenAI", new_callable=MagicMock)
@patch(
    "mirascope.core.openai._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.openai._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_stream(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_openai: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_utils.setup_call = mock_base_setup_call
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=True,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "model" in call_kwargs and call_kwargs["model"] == "gpt-4o"
    assert "messages" in call_kwargs and call_kwargs["messages"] == messages
    assert "stream_options" in call_kwargs and call_kwargs["stream_options"] == {
        "include_usage": True
    }
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, OpenAITool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1]
    )
    assert messages == mock_convert_message_params.return_value
    mock_create = mock_openai.return_value.chat.completions.create
    assert create(stream=True, **call_kwargs)
    mock_create.assert_called_once_with(**call_kwargs, stream=True)


@patch(
    "mirascope.core.openai._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.openai._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock()
    mock_utils.json_mode_content.return_value = "\n\njson output"
    mock_base_setup_call.return_value[1] = [
        {"role": "user", "content": [{"type": "text", "text": "test"}]}
    ]

    class MockToolType(MagicMock):
        model_config: ClassVar[dict] = {}

    mock_base_setup_call.return_value[2] = [MockToolType]
    mock_base_setup_call.return_value[-1]["tools"] = MagicMock()
    mock_convert_message_params.side_effect = lambda x: x
    _, _, messages, _, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert messages[-1] == {"role": "user", "content": "json output"}
    assert "tools" in call_kwargs

    mock_base_setup_call.return_value[1] = [
        {"role": "assistant", "content": [{"type": "text", "text": "test"}]}
    ]
    _, _, messages, _, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert isinstance(messages[-1], dict) and "content" in messages[-1]

    assert messages[-1] == {"role": "user", "content": "json output"}

    class Tool(OpenAITool):
        """A test tool."""

        title: str

        model_config = ResponseModelConfigDict(strict=True)

    mock_base_setup_call.return_value[2] = [Tool]
    _, _, _, _, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=Tool,
        stream=False,
    )
    assert "response_format" in call_kwargs and call_kwargs["response_format"] == {
        "type": "json_schema",
        "json_schema": {
            "name": "Tool",
            "description": "A test tool.",
            "strict": True,
            "schema": {
                "properties": {"title": {"type": "string"}},
                "required": ["title"],
                "type": "object",
                "additionalProperties": False,
            },
        },
    }


@patch(
    "mirascope.core.openai._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.openai._utils._setup_call._utils", new_callable=MagicMock)
def test_setup_call_extract(
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""

    class Tool(OpenAITool):
        """A test tool."""

        title: str

        model_config = ResponseModelConfigDict(strict=True)

    mock_base_setup_call.return_value[2] = [Tool]
    mock_utils.setup_call = mock_base_setup_call
    with pytest.warns(
        UserWarning,
        match="You must set `json_mode=True` to use `strict=True` response models. "
        "Ignoring `strict` and using tools for extraction.",
    ):
        _, _, _, _, call_kwargs = setup_call(
            model="gpt-4o",
            client=None,
            fn=MagicMock(),
            fn_args={},
            dynamic_config=None,
            tools=None,
            json_mode=False,
            call_params={},
            response_model=BaseModel,
            stream=False,
        )
    assert "tool_choice" in call_kwargs and call_kwargs["tool_choice"] == "required"
